import pytest

import salt.utils.win_lgpo_netsh as win_lgpo_netsh
from salt.exceptions import CommandExecutionError
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


def test_get_settings_firewallpolicy_local():
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="local"
    )
    assert "Inbound" in ret
    assert "Outbound" in ret


def test_get_settings_firewallpolicy_lgpo():
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="lgpo"
    )
    assert "Inbound" in ret
    assert "Outbound" in ret


def test_get_settings_firewallpolicy_lgpo_issue_57591():
    """
    Should not stacktrace when the hostname contains unicode characters
    """
    with patch.object(win_lgpo_netsh, "__hostname__", return_value="kомпьютер"):
        ret = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="lgpo"
        )
    assert "Inbound" in ret
    assert "Outbound" in ret


def test_get_settings_logging_local():
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret


def test_get_settings_logging_lgpo():
    ret = win_lgpo_netsh.get_settings(profile="domain", section="logging", store="lgpo")
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret


def test_get_settings_settings_local():
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="local"
    )
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "RemoteManagement" in ret
    assert "UnicastResponseToMulticast" in ret


def test_get_settings_settings_lgpo():
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="lgpo"
    )
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "RemoteManagement" in ret
    assert "UnicastResponseToMulticast" in ret


def test_get_settings_state_local():
    ret = win_lgpo_netsh.get_settings(profile="domain", section="state", store="local")
    assert "State" in ret


def test_get_settings_state_lgpo():
    ret = win_lgpo_netsh.get_settings(profile="domain", section="state", store="lgpo")
    assert "State" in ret


def test_get_all_settings_local():
    ret = win_lgpo_netsh.get_all_settings(profile="domain", store="local")
    assert "Inbound" in ret
    assert "Outbound" in ret
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "RemoteManagement" in ret
    assert "UnicastResponseToMulticast" in ret
    assert "State" in ret


def test_get_all_settings_lgpo():
    ret = win_lgpo_netsh.get_all_settings(profile="domain", store="local")
    assert "Inbound" in ret
    assert "Outbound" in ret
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "RemoteManagement" in ret
    assert "UnicastResponseToMulticast" in ret
    assert "State" in ret


def test_get_all_profiles_local():
    ret = win_lgpo_netsh.get_all_profiles(store="local")
    assert "Domain Profile" in ret
    assert "Private Profile" in ret
    assert "Public Profile" in ret


def test_get_all_profiles_lgpo():
    ret = win_lgpo_netsh.get_all_profiles(store="lgpo")
    assert "Domain Profile" in ret
    assert "Private Profile" in ret
    assert "Public Profile" in ret


@pytest.mark.destructive_test
def test_set_firewall_settings_inbound_local():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="local"
    )["Inbound"]
    try:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", inbound="allowinbound", store="local"
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )["Inbound"]
        assert new == "AllowInbound"
    finally:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", inbound=current, store="local"
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_inbound_local_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="local"
    )["Inbound"]
    try:
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_firewall_settings,
            profile="domain",
            inbound="notconfigured",
            store="local",
        )
    finally:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", inbound=current, store="local"
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_inbound_lgpo_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="lgpo"
    )["Inbound"]
    try:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", inbound="notconfigured", store="lgpo"
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="lgpo"
        )["Inbound"]
        assert new == "NotConfigured"
    finally:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", inbound=current, store="lgpo"
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_outbound_local():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store="local"
    )["Outbound"]
    try:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", outbound="allowoutbound", store="local"
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store="local"
        )["Outbound"]
        assert new == "AllowOutbound"
    finally:
        ret = win_lgpo_netsh.set_firewall_settings(
            profile="domain", outbound=current, store="local"
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_logging_allowed_local_enable():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )["LogAllowedConnections"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="allowedconnections",
            value="enable",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["LogAllowedConnections"]
        assert new == "Enable"
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="allowedconnections",
            value=current,
            store="local",
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_logging_allowed_local_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )["LogAllowedConnections"]
    try:
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_logging_settings,
            profile="domain",
            setting="allowedconnections",
            value="notconfigured",
            store="local",
        )
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="allowedconnections",
            value=current,
            store="local",
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_logging_allowed_lgpo_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="lgpo"
    )["LogAllowedConnections"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="allowedconnections",
            value="notconfigured",
            store="lgpo",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="lgpo"
        )["LogAllowedConnections"]
        assert new == "NotConfigured"
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="allowedconnections",
            value=current,
            store="lgpo",
        )
        assert ret is True


def test_set_firewall_logging_dropped_local_enable():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )["LogDroppedConnections"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="droppedconnections",
            value="enable",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["LogDroppedConnections"]
        assert new == "Enable"
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="droppedconnections",
            value=current,
            store="local",
        )
        assert ret is True


def test_set_firewall_logging_filename_local():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )["FileName"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="filename",
            value="C:\\Temp\\test.log",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["FileName"]
        assert new == "C:\\Temp\\test.log"
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain", setting="filename", value=current, store="local"
        )
        assert ret is True


def test_set_firewall_logging_maxfilesize_local():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store="local"
    )["MaxFileSize"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain", setting="maxfilesize", value="16384", store="local"
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store="local"
        )["MaxFileSize"]
        assert new == "16384"
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain", setting="maxfilesize", value=current, store="local"
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_fwrules_local_enable():
    pytest.raises(
        CommandExecutionError,
        win_lgpo_netsh.set_settings,
        profile="domain",
        setting="localfirewallrules",
        value="enable",
        store="local",
    )


@pytest.mark.destructive_test
def test_set_firewall_settings_fwrules_lgpo_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="lgpo"
    )["LocalFirewallRules"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="localfirewallrules",
            value="notconfigured",
            store="lgpo",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="lgpo"
        )["LocalFirewallRules"]
        assert new == "NotConfigured"
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="localfirewallrules",
            value=current,
            store="lgpo",
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_consecrules_local_enable():
    pytest.raises(
        CommandExecutionError,
        win_lgpo_netsh.set_settings,
        profile="domain",
        setting="localconsecrules",
        value="enable",
        store="local",
    )


def test_set_firewall_settings_notification_local_enable():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="local"
    )["InboundUserNotification"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="inboundusernotification",
            value="enable",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["InboundUserNotification"]
        assert new == "Enable"
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="inboundusernotification",
            value=current,
            store="local",
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_settings_notification_local_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="local"
    )["InboundUserNotification"]
    try:
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_settings,
            profile="domain",
            setting="inboundusernotification",
            value="notconfigured",
            store="local",
        )
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="inboundusernotification",
            value=current,
            store="local",
        )
        assert ret is True


def test_set_firewall_settings_notification_lgpo_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="lgpo"
    )["InboundUserNotification"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="inboundusernotification",
            value="notconfigured",
            store="lgpo",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="lgpo"
        )["InboundUserNotification"]
        assert new == "NotConfigured"
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="inboundusernotification",
            value=current,
            store="lgpo",
        )
        assert ret is True


def test_set_firewall_settings_remotemgmt_local_enable():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="local"
    )["RemoteManagement"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="remotemanagement",
            value="enable",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["RemoteManagement"]
        assert new == "Enable"
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="remotemanagement",
            value=current,
            store="local",
        )
        assert ret is True


def test_set_firewall_settings_unicast_local_disable():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="local"
    )["UnicastResponseToMulticast"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="unicastresponsetomulticast",
            value="disable",
            store="local",
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store="local"
        )["UnicastResponseToMulticast"]
        assert new == "Disable"
    finally:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="unicastresponsetomulticast",
            value=current,
            store="local",
        )
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_state_local_on():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="state", store="local"
    )["State"]
    try:
        ret = win_lgpo_netsh.set_state(profile="domain", state="off", store="local")
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="local"
        )["State"]
        assert new == "OFF"
    finally:
        ret = win_lgpo_netsh.set_state(profile="domain", state=current, store="local")
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_state_local_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="state", store="local"
    )["State"]
    try:
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_state,
            profile="domain",
            state="notconfigured",
            store="local",
        )
    finally:
        ret = win_lgpo_netsh.set_state(profile="domain", state=current, store="local")
        assert ret is True


@pytest.mark.destructive_test
def test_set_firewall_state_lgpo_notconfigured():
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="state", store="local"
    )["State"]
    try:
        ret = win_lgpo_netsh.set_state(
            profile="domain", state="notconfigured", store="lgpo"
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store="lgpo"
        )["State"]
        assert new == "NotConfigured"
    finally:
        ret = win_lgpo_netsh.set_state(profile="domain", state=current, store="lgpo")
        assert ret is True



"""
Management of Docker images

.. versionadded:: 2017.7.0

:depends: docker_ Python module

.. note::
    Older releases of the Python bindings for Docker were called docker-py_ in
    PyPI. All releases of docker_, and releases of docker-py_ >= 1.6.0 are
    supported. These python bindings can easily be installed using
    :py:func:`pip.install <salt.modules.pip.install>`:

    .. code-block:: bash

        salt myminion pip.install docker

    To upgrade from docker-py_ to docker_, you must first uninstall docker-py_,
    and then install docker_:

    .. code-block:: bash

        salt myminion pip.uninstall docker-py
        salt myminion pip.install docker

.. _docker: https://pypi.python.org/pypi/docker
.. _docker-py: https://pypi.python.org/pypi/docker-py

These states were moved from the :mod:`docker <salt.states.docker>` state
module (formerly called **dockerng**) in the 2017.7.0 release.

.. note::
    To pull from a Docker registry, authentication must be configured. See
    :ref:`here <docker-authentication>` for more information on how to
    configure access to docker registries in :ref:`Pillar <pillar>` data.
"""

import logging

import salt.utils.args
import salt.utils.dockermod
from salt.exceptions import CommandExecutionError

# Enable proper logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "docker_image"
__virtual_aliases__ = ("moby_image",)


def __virtual__():
    """
    Only load if the docker execution module is available
    """
    if "docker.version" in __salt__:
        return __virtualname__
    return (False, __salt__.missing_fun_string("docker.version"))


def present(
    name,
    tag=None,
    build=None,
    load=None,
    force=False,
    insecure_registry=False,
    client_timeout=salt.utils.dockermod.CLIENT_TIMEOUT,
    dockerfile=None,
    sls=None,
    base="opensuse/python",
    saltenv="base",
    pillarenv=None,
    pillar=None,
    **kwargs
):
    """
    .. versionchanged:: 2018.3.0
        The ``tag`` argument has been added. It is now required unless pulling
        from a registry.

    Ensure that an image is present. The image can either be pulled from a
    Docker registry, built from a Dockerfile, loaded from a saved image, or
    built by running SLS files against a base image.

    If none of the ``build``, ``load``, or ``sls`` arguments are used, then Salt
    will pull from the :ref:`configured registries <docker-authentication>`. If
    the specified image already exists, it will not be pulled unless ``force``
    is set to ``True``. Here is an example of a state that will pull an image
    from the Docker Hub:

    .. code-block:: yaml

        myuser/myimage:
          docker_image.present:
            - tag: mytag

    tag
        Tag name for the image. Required when using ``build``, ``load``, or
        ``sls`` to create the image, but optional if pulling from a repository.

        .. versionadded:: 2018.3.0

    build
        Path to directory on the Minion containing a Dockerfile

        .. code-block:: yaml

            myuser/myimage:
              docker_image.present:
                - build: /home/myuser/docker/myimage
                - tag: mytag

            myuser/myimage:
              docker_image.present:
                - build: /home/myuser/docker/myimage
                - tag: mytag
                - dockerfile: Dockerfile.alternative

        The image will be built using :py:func:`docker.build
        <salt.modules.dockermod.build>` and the specified image name and tag
        will be applied to it.

        .. versionadded:: 2016.11.0
        .. versionchanged:: 2018.3.0
            The ``tag`` must be manually specified using the ``tag`` argument.

    load
        Loads a tar archive created with :py:func:`docker.save
        <salt.modules.dockermod.save>` (or the ``docker save`` Docker CLI
        command), and assigns it the specified repo and tag.

        .. code-block:: yaml

            myuser/myimage:
              docker_image.present:
                - load: salt://path/to/image.tar
                - tag: mytag

        .. versionchanged:: 2018.3.0
            The ``tag`` must be manually specified using the ``tag`` argument.

    force : False
        Set this parameter to ``True`` to force Salt to pull/build/load the
        image even if it is already present.

    client_timeout
        Timeout in seconds for the Docker client. This is not a timeout for
        the state, but for receiving a response from the API.

    dockerfile
        Allows for an alternative Dockerfile to be specified.  Path to alternative
        Dockefile is relative to the build path for the Docker container.

        .. versionadded:: 2016.11.0

    sls
        Allow for building of image with :py:func:`docker.sls_build
        <salt.modules.dockermod.sls_build>` by specifying the SLS files with
        which to build. This can be a list or comma-separated string.

        .. code-block:: yaml

            myuser/myimage:
              docker_image.present:
                - tag: latest
                - sls:
                    - webapp1
                    - webapp2
                - base: centos
                - saltenv: base

        .. versionadded:: 2017.7.0
        .. versionchanged:: 2018.3.0
            The ``tag`` must be manually specified using the ``tag`` argument.

    base
        Base image with which to start :py:func:`docker.sls_build
        <salt.modules.dockermod.sls_build>`

        .. versionadded:: 2017.7.0

    saltenv
        Specify the environment from which to retrieve the SLS indicated by the
        `mods` parameter.

        .. versionadded:: 2017.7.0
        .. versionchanged:: 2018.3.0
            Now uses the effective saltenv if not explicitly passed. In earlier
            versions, ``base`` was assumed as a default.

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

        .. versionadded:: 2018.3.0

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. note::
            Values passed this way will override Pillar values set via
            ``pillar_roots`` or an external Pillar source.

        .. versionadded:: 2018.3.0
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if not isinstance(name, str):
        name = str(name)

    # At most one of the args that result in an image being built can be used
    num_build_args = len([x for x in (build, load, sls) if x is not None])
    if num_build_args > 1:
        ret["comment"] = "Only one of 'build', 'load', or 'sls' is permitted."
        return ret
    elif num_build_args == 1:
        # If building, we need the tag to be specified
        if not tag:
            ret["comment"] = (
                "The 'tag' argument is required if any one of 'build', "
                "'load', or 'sls' is used."
            )
            return ret
        if not isinstance(tag, str):
            tag = str(tag)
        full_image = ":".join((name, tag))
    else:
        if tag:
            name = "{}:{}".format(name, tag)
        full_image = name

    try:
        image_info = __salt__["docker.inspect_image"](full_image)
    except CommandExecutionError as exc:
        msg = exc.__str__()
        if "404" in msg:
            # Image not present
            image_info = None
        else:
            ret["comment"] = msg
            return ret

    if image_info is not None:
        # Specified image is present
        if not force:
            ret["result"] = True
            ret["comment"] = "Image {} already present".format(full_image)
            return ret

    if build or sls:
        action = "built"
    elif load:
        action = "loaded"
    else:
        action = "pulled"

    if __opts__["test"]:
        ret["result"] = None
        if (image_info is not None and force) or image_info is None:
            ret["comment"] = "Image {} will be {}".format(full_image, action)
            return ret

    if build:
        # Get the functions default value and args
        argspec = salt.utils.args.get_function_argspec(__salt__["docker.build"])
        # Map any if existing args from kwargs into the build_args dictionary
        build_args = dict(list(zip(argspec.args, argspec.defaults)))
        for k in build_args:
            if k in kwargs.get("kwargs", {}):
                build_args[k] = kwargs.get("kwargs", {}).get(k)
        try:
            # map values passed from the state to the build args
            build_args["path"] = build
            build_args["repository"] = name
            build_args["tag"] = tag
            build_args["dockerfile"] = dockerfile
            image_update = __salt__["docker.build"](**build_args)
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Encountered error building {} as {}: {}".format(
                build, full_image, exc
            )
            return ret
        if image_info is None or image_update["Id"] != image_info["Id"][:12]:
            ret["changes"] = image_update

    elif sls:
        _locals = locals()
        sls_build_kwargs = {
            k: _locals[k]
            for k in ("saltenv", "pillarenv", "pillar")
            if _locals[k] is not None
        }
        try:
            image_update = __salt__["docker.sls_build"](
                repository=name, tag=tag, base=base, mods=sls, **sls_build_kwargs
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret[
                "comment"
            ] = "Encountered error using SLS {} for building {}: {}".format(
                sls, full_image, exc
            )
            return ret
        if image_info is None or image_update["Id"] != image_info["Id"][:12]:
            ret["changes"] = image_update

    elif load:
        try:
            image_update = __salt__["docker.load"](path=load, repository=name, tag=tag)
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Encountered error loading {} as {}: {}".format(
                load, full_image, exc
            )
            return ret
        if image_info is None or image_update.get("Layers", []):
            ret["changes"] = image_update

    else:
        try:
            image_update = __salt__["docker.pull"](
                name, insecure_registry=insecure_registry, client_timeout=client_timeout
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Encountered error pulling {}: {}".format(full_image, exc)
            return ret
        if (
            image_info is not None
            and image_info["Id"][:12]
            == image_update.get("Layers", {}).get("Already_Pulled", [None])[0]
        ):
            # Image was pulled again (because of force) but was also
            # already there. No new image was available on the registry.
            pass
        elif image_info is None or image_update.get("Layers", {}).get("Pulled"):
            # Only add to the changes dict if layers were pulled
            ret["changes"] = image_update

    error = False

    try:
        __salt__["docker.inspect_image"](full_image)
    except CommandExecutionError as exc:
        msg = exc.__str__()
        if "404" not in msg:
            error = "Failed to inspect image '{}' after it was {}: {}".format(
                full_image, action, msg
            )

    if error:
        ret["comment"] = error
    else:
        ret["result"] = True
        if not ret["changes"]:
            ret["comment"] = "Image '{}' was {}, but there were no changes".format(
                name, action
            )
        else:
            ret["comment"] = "Image '{}' was {}".format(full_image, action)
    return ret


def absent(name=None, images=None, force=False):
    """
    Ensure that an image is absent from the Minion. Image names can be
    specified either using ``repo:tag`` notation, or just the repo name (in
    which case a tag of ``latest`` is assumed).

    images
        Run this state on more than one image at a time. The following two
        examples accomplish the same thing:

        .. code-block:: yaml

            remove_images:
              docker_image.absent:
                - names:
                  - busybox
                  - centos:6
                  - nginx

        .. code-block:: yaml

            remove_images:
              docker_image.absent:
                - images:
                  - busybox
                  - centos:6
                  - nginx

        However, the second example will be a bit quicker since Salt will do
        all the deletions in a single run, rather than executing the state
        separately on each image (as it would in the first example).

    force : False
        Salt will fail to remove any images currently in use by a container.
        Set this option to true to remove the image even if it is already
        present.

        .. note::

            This option can also be overridden by Pillar data. If the Minion
            has a pillar variable named ``docker.running.force`` which is
            set to ``True``, it will turn on this option. This pillar variable
            can even be set at runtime. For example:

            .. code-block:: bash

                salt myminion state.sls docker_stuff pillar="{docker.force: True}"

            If this pillar variable is present and set to ``False``, then it
            will turn off this option.

            For more granular control, setting a pillar variable named
            ``docker.force.image_name`` will affect only the named image.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if not name and not images:
        ret["comment"] = "One of 'name' and 'images' must be provided"
        return ret
    elif images is not None:
        targets = images
    elif name:
        targets = [name]

    to_delete = []
    for target in targets:
        resolved_tag = __salt__["docker.resolve_tag"](target)
        if resolved_tag is not False:
            to_delete.append(resolved_tag)

    if not to_delete:
        ret["result"] = True
        if len(targets) == 1:
            ret["comment"] = "Image {} is not present".format(name)
        else:
            ret["comment"] = "All specified images are not present"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        if len(to_delete) == 1:
            ret["comment"] = "Image {} will be removed".format(to_delete[0])
        else:
            ret["comment"] = "The following images will be removed: {}".format(
                ", ".join(to_delete)
            )
        return ret

    result = __salt__["docker.rmi"](*to_delete, force=force)
    post_tags = __salt__["docker.list_tags"]()
    failed = [x for x in to_delete if x in post_tags]

    if failed:
        if [x for x in to_delete if x not in post_tags]:
            ret["changes"] = result
            ret["comment"] = "The following image(s) failed to be removed: {}".format(
                ", ".join(failed)
            )
        else:
            ret["comment"] = "None of the specified images were removed"
            if "Errors" in result:
                ret["comment"] += ". The following errors were encountered: {}".format(
                    "; ".join(result["Errors"])
                )
    else:
        ret["changes"] = result
        if len(to_delete) == 1:
            ret["comment"] = "Image {} was removed".format(to_delete[0])
        else:
            ret["comment"] = "The following images were removed: {}".format(
                ", ".join(to_delete)
            )
        ret["result"] = True

    return ret


def mod_watch(name, sfun=None, **kwargs):
    """
    The docker_image  watcher, called to invoke the watch command.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.
    """
    if sfun == "present":
        # Force image to be updated
        kwargs["force"] = True
        return present(name, **kwargs)

    return {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "watch requisite is not implemented for {}".format(sfun),
    }



"""
tests.unit.modules.test_glassfish
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the glassfish module
"""
import logging

import pytest

import salt.modules.glassfish as glassfish
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {glassfish: {}}


def test__api_get():
    get_mock = MagicMock()
    with patch("salt.modules.glassfish.requests.get", get_mock):
        glassfish._api_get("ThePath", server=glassfish.DEFAULT_SERVER)

    get_mock.assert_called_once_with(
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Requested-By": "GlassFish REST HTML interface",
        },
        url="http://localhost:4848/management/domain/ThePath",
        verify=True,
        auth=None,
    )


def test__api_post():
    post_mock = MagicMock()
    with patch("salt.modules.glassfish.requests.post", post_mock):
        glassfish._api_post("ThePath", {1: 1}, server=glassfish.DEFAULT_SERVER)

    post_mock.assert_called_once_with(
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Requested-By": "GlassFish REST HTML interface",
        },
        url="http://localhost:4848/management/domain/ThePath",
        verify=True,
        auth=None,
        data='{"1": 1}',
    )


def test__api_delete():
    delete_mock = MagicMock()
    with patch("salt.modules.glassfish.requests.delete", delete_mock):
        glassfish._api_delete("ThePath", {1: 1}, server=glassfish.DEFAULT_SERVER)

    delete_mock.assert_called_once_with(
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Requested-By": "GlassFish REST HTML interface",
        },
        url="http://localhost:4848/management/domain/ThePath",
        verify=True,
        auth=None,
        params={1: 1},
    )



"""
    :codeauthor: :email:`Shane Lee <slee@saltstack.com>`
"""

import pytest

import salt.grains.lvm as lvm
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        lvm: {"__salt__": {}},
    }


def test__linux_lvm():
    """
    Test grains._linux_lvm, normal return
    Should return a populated dictionary
    """

    vgs_out = {"pid": 123, "retcode": 0, "stdout": "  vg00\n  vg01", "stderr": ""}
    lvs_out_vg00 = {
        "pid": 456,
        "retcode": 0,
        "stdout": "  root\n  swap\n  tmp \n  usr \n  var",
        "stderr": "",
    }
    lvs_out_vg01 = {"pid": 789, "retcode": 0, "stdout": "  opt", "stderr": ""}
    cmd_out = MagicMock(
        autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
    )

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret == {
        "lvm": {"vg00": ["root", "swap", "tmp", "usr", "var"], "vg01": ["opt"]}
    }, ret


def test__linux_lvm_with_WARNINGs():
    """
    Test grains._linux_lvm, with WARNINGs in lvm command output
    Should return a populated dictionary
    """

    vgs_out = {
        "pid": 123,
        "retcode": 0,
        "stdout": "  vg00\n  vg01",
        "stderr": "WARNING: Something wrong is not right",
    }
    lvs_out_vg00 = {
        "pid": 456,
        "retcode": 0,
        "stdout": "  root\n  swap\n  tmp \n  usr \n  var",
        "stderr": "WARNING: Something wrong is not right",
    }
    lvs_out_vg01 = {
        "pid": 789,
        "retcode": 0,
        "stdout": "  opt",
        "stderr": "WARNING: Something wrong is not right",
    }
    cmd_out = MagicMock(
        autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
    )

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret == {
        "lvm": {"vg00": ["root", "swap", "tmp", "usr", "var"], "vg01": ["opt"]}
    }, ret


def test__linux_lvm_with_non_zero_exit_codes():
    """
    Test grains._linux_lvm, with non-zero exit codes for lvm command
    Should return a populated dictionary
    """

    vgs_out = {
        "pid": 123,
        "retcode": 5,
        "stdout": "  vg00\n  vg01",
        "stderr": "  Skipping clustered volume vgcluster\n  Skipping volume group vgcluster",
    }
    lvs_out_vg00 = {
        "pid": 456,
        "retcode": 0,
        "stdout": "  root\n  swap\n  tmp \n  usr \n  var",
        "stderr": "",
    }
    lvs_out_vg01 = {"pid": 789, "retcode": 0, "stdout": "  opt", "stderr": ""}
    cmd_out = MagicMock(
        autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
    )

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret == {
        "lvm": {"vg00": ["root", "swap", "tmp", "usr", "var"], "vg01": ["opt"]}
    }, ret


def test__linux_lvm_no_lvm():
    """
    Test grains._linux_lvm, no lvm installed
    Should return nothing
    """

    vgs_out = {"pid": 123, "retcode": 0, "stdout": "  vg00\n  vg01", "stderr": ""}
    lvs_out_vg00 = {
        "pid": 456,
        "retcode": 0,
        "stdout": "  root\n  swap\n  tmp \n  usr \n  var",
        "stderr": "",
    }
    lvs_out_vg01 = {"pid": 789, "retcode": 0, "stdout": "  opt", "stderr": ""}
    cmd_out = MagicMock(
        autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
    )

    patch_which = patch("salt.utils.path.which", autospec=True, return_value="")
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret is None, ret


def test__linux_lvm_no_volume_groups():
    """
    Test grains._linux_lvm, lvm is installed but no volume groups created.
    Should return a dictionary only with the header
    """

    vgs_out = {"pid": 123, "retcode": 0, "stdout": "", "stderr": ""}
    cmd_out = MagicMock(autospec=True, side_effect=[vgs_out])

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret == {"lvm": {}}, ret


def test__linux_lvm_no_logical_volumes():
    """
    Test grains._linux_lvm, lvm is installed, volume groups created but
    no logical volumes present.
    Should return a dictionary only with the header
    """

    vgs_out = {"pid": 123, "retcode": 0, "stdout": "  vg00\n  vg01", "stderr": ""}
    lvs_out = {"pid": 456, "retcode": 0, "stdout": "", "stderr": ""}
    cmd_out = MagicMock(autospec=True, side_effect=[vgs_out, lvs_out, lvs_out])

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run_all": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._linux_lvm()

    assert ret == {"lvm": {"vg00": [], "vg01": []}}, ret


def test__aix_lvm():
    """
    Test grains._aix_lvm, normal return
    Should return a populated dictionary
    """

    lsvg_out = "rootvg\nothervg"
    lsvg_out_rootvg = (
        "rootvg:\nLV NAME             TYPE       LPs     PPs     PVs  LV STATE     "
        " MOUNT POINT\nhd5                 boot       1       1       1   "
        " closed/syncd  N/A\nhd6                 paging     32      32      1   "
        " open/syncd    N/A\nhd8                 jfs2log    1       1       1   "
        " open/syncd    N/A\nhd4                 jfs2       32      32      1   "
        " open/syncd    /\nhd2                 jfs2       16      16      1   "
        " open/syncd    /usr\nhd9var              jfs2       32      32      1   "
        " open/syncd    /var\nhd3                 jfs2       32      32      1   "
        " open/syncd    /tmp\nhd1                 jfs2       16      16      1   "
        " open/syncd    /home\nhd10opt             jfs2       16      16      1   "
        " open/syncd    /opt"
    )
    lsvg_out_othervg = (
        "othervg:\nLV NAME             TYPE       LPs     PPs     PVs  LV STATE    "
        "  MOUNT POINT\nloglv01             jfs2log    1       1       1   "
        " open/syncd    N/A\ndatalv              jfs2       16      16      1   "
        " open/syncd    /data"
    )
    cmd_out = MagicMock(
        autospec=True, side_effect=[lsvg_out, lsvg_out_rootvg, lsvg_out_othervg]
    )

    patch_which = patch(
        "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lsvg"
    )
    patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run": cmd_out})
    with patch_which, patch_cmd_lvm:
        ret = lvm._aix_lvm()

    assert ret == {
        "lvm": {
            "rootvg": [
                "hd5",
                "hd6",
                "hd8",
                "hd4",
                "hd2",
                "hd9var",
                "hd3",
                "hd1",
                "hd10opt",
            ],
            "othervg": ["loglv01", "datalv"],
        }
    }, ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import pytest

import salt.states.slack as slack
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {slack: {}}


def test_post_message_apikey():
    """
    Test to send a message to a Slack channel using an API Key.
    """
    name = "slack-message"
    channel = "#general"
    from_name = "SuperAdmin"
    message = "This state was executed successfully."
    api_key = "xoxp-XXXXXXXXXX-XXXXXXXXXX-XXXXXXXXXX-XXXXXX"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(slack.__opts__, {"test": True}):
        mock = MagicMock(return_value=True)
        with patch.dict(slack.__salt__, {"config.get": mock}):
            comt = "The following message is to be sent to Slack: {}".format(message)
            ret.update({"comment": comt})
            assert (
                slack.post_message(
                    name,
                    channel=channel,
                    from_name=from_name,
                    message=message,
                    api_key=api_key,
                )
                == ret
            )

    with patch.dict(slack.__opts__, {"test": False}):
        mock = MagicMock(return_value=False)
        with patch.dict(slack.__salt__, {"config.get": mock}):
            comt = "Please specify api_key or webhook."
            ret.update({"comment": comt, "result": False})
            assert (
                slack.post_message(
                    name,
                    channel=None,
                    from_name=from_name,
                    message=message,
                    api_key=None,
                )
                == ret
            )

        comt = "Slack channel is missing."
        ret.update({"comment": comt, "result": False})
        assert (
            slack.post_message(
                name,
                channel=None,
                from_name=from_name,
                message=message,
                api_key=api_key,
            )
            == ret
        )

        comt = "Slack from name is missing."
        ret.update({"comment": comt, "result": False})
        assert (
            slack.post_message(
                name,
                channel=channel,
                from_name=None,
                message=message,
                api_key=api_key,
            )
            == ret
        )

        comt = "Slack message is missing."
        ret.update({"comment": comt, "result": False})
        assert (
            slack.post_message(
                name,
                channel=channel,
                from_name=from_name,
                message=None,
                api_key=api_key,
            )
            == ret
        )

        mock = MagicMock(return_value=True)
        with patch.dict(slack.__salt__, {"slack.post_message": mock}):
            comt = "Sent message: slack-message"
            ret.update({"comment": comt, "result": True})
            assert (
                slack.post_message(
                    name,
                    channel=channel,
                    from_name=from_name,
                    message=message,
                    api_key=api_key,
                )
                == ret
            )


def test_post_message_webhook():
    """
    Test to send a message to a Slack channel using an webhook.
    """
    name = "slack-message"
    channel = "#general"
    username = "SuperAdmin"
    message = "This state was executed successfully."
    webhook = "XXXXXXXXX/XXXXXXXXX/XXXXXXXXXXXXXXXXXXXXXXXX"
    api_key = "xoxp-XXXXXXXXXX-XXXXXXXXXX-XXXXXXXXXX-XXXXXX"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(slack.__opts__, {"test": True}):
        mock = MagicMock(return_value=True)
        with patch.dict(slack.__salt__, {"config.get": mock}):
            comt = "The following message is to be sent to Slack: {}".format(message)
            ret.update({"comment": comt})
            assert (
                slack.post_message(
                    name,
                    channel=channel,
                    username=username,
                    message=message,
                    webhook=webhook,
                )
                == ret
            )

    with patch.dict(slack.__opts__, {"test": False}):
        mock = MagicMock(return_value=False)
        with patch.dict(slack.__salt__, {"config.get": mock}):
            comt = "Please specify api_key or webhook."
            ret.update({"comment": comt, "result": False})
            assert (
                slack.post_message(
                    name, channel=channel, username=username, message=None, webhook=None
                )
                == ret
            )

        comt = "Please specify only either api_key or webhook."
        ret.update({"comment": comt, "result": False})
        assert (
            slack.post_message(
                name,
                channel=channel,
                username=username,
                message=message,
                api_key=api_key,
                webhook=webhook,
            )
            == ret
        )

        comt = "Slack message is missing."
        ret.update({"comment": comt, "result": False})
        assert (
            slack.post_message(
                name,
                channel=channel,
                username=username,
                message=None,
                webhook=webhook,
            )
            == ret
        )

        mock = MagicMock(return_value=True)
        with patch.dict(slack.__salt__, {"slack.call_hook": mock}):
            comt = "Sent message: slack-message"
            ret.update({"comment": comt, "result": True})
            assert (
                slack.post_message(
                    name,
                    channel=channel,
                    username=username,
                    message=message,
                    webhook=webhook,
                )
                == ret
            )



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    saltautodoc.py
    ~~~~~~~~~~~~~~

    Properly handle ``__func_alias__``
"""

from sphinx.ext.autodoc import FunctionDocumenter


class SaltFunctionDocumenter(FunctionDocumenter):
    """
    Simple override of sphinx.ext.autodoc.FunctionDocumenter to properly render
    salt's aliased function names.
    """

    def format_name(self):
        """
        Format the function name
        """
        if not hasattr(self.module, "__func_alias__"):
            # Resume normal sphinx.ext.autodoc operation
            return super(FunctionDocumenter, self).format_name()

        if not self.objpath:
            # Resume normal sphinx.ext.autodoc operation
            return super(FunctionDocumenter, self).format_name()

        if len(self.objpath) > 1:
            # Resume normal sphinx.ext.autodoc operation
            return super(FunctionDocumenter, self).format_name()

        # Use the salt func aliased name instead of the real name
        return self.module.__func_alias__.get(self.objpath[0], self.objpath[0])


def setup(app):
    def add_documenter(app, env, docnames):
        app.add_autodocumenter(SaltFunctionDocumenter)

    # add_autodocumenter() must be called after the initial setup and the
    # 'builder-inited' event, as sphinx.ext.autosummary will restore the
    # original documenter on 'builder-inited'
    app.connect("env-before-read-docs", add_documenter)



"""
Many aspects of the salt payload need to be managed, from the return of
encrypted keys to general payload dynamics and packaging, these happen
in here
"""

import collections.abc
import datetime
import gc
import logging

import salt.loader.context
import salt.transport.frame
import salt.utils.immutabletypes as immutabletypes
import salt.utils.msgpack
import salt.utils.stringutils
from salt.defaults import _Constant
from salt.exceptions import SaltDeserializationError, SaltReqTimeoutError
from salt.utils.data import CaseInsensitiveDict

try:
    import zmq
except ImportError:
    # No need for zeromq in local mode
    pass

log = logging.getLogger(__name__)


def package(payload):
    """
    This method for now just wraps msgpack.dumps, but it is here so that
    we can make the serialization a custom option in the future with ease.
    """
    return salt.utils.msgpack.dumps(payload)


def unpackage(package_):
    """
    Unpackages a payload
    """
    return salt.utils.msgpack.loads(package_, use_list=True)


def format_payload(enc, **kwargs):
    """
    Pass in the required arguments for a payload, the enc type and the cmd,
    then a list of keyword args to generate the body of the load dict.
    """
    payload = {"enc": enc}
    load = {}
    for key in kwargs:
        load[key] = kwargs[key]
    payload["load"] = load
    return package(payload)


def loads(msg, encoding=None, raw=False):
    """
    Run the correct loads serialization format

    :param encoding: Useful for Python 3 support. If the msgpack data
                     was encoded using "use_bin_type=True", this will
                     differentiate between the 'bytes' type and the
                     'str' type by decoding contents with 'str' type
                     to what the encoding was set as. Recommended
                     encoding is 'utf-8' when using Python 3.
                     If the msgpack data was not encoded using
                     "use_bin_type=True", it will try to decode
                     all 'bytes' and 'str' data (the distinction has
                     been lost in this case) to what the encoding is
                     set as. In this case, it will fail if any of
                     the contents cannot be converted.
    """
    try:

        def ext_type_decoder(code, data):
            if code == 78:
                data = salt.utils.stringutils.to_unicode(data)
                return datetime.datetime.strptime(data, "%Y%m%dT%H:%M:%S.%f")
            if code == 79:
                name, value = salt.utils.msgpack.loads(data, raw=False)
                return _Constant(name, value)
            return data

        gc.disable()  # performance optimization for msgpack
        loads_kwargs = {"use_list": True, "ext_hook": ext_type_decoder}
        if salt.utils.msgpack.version >= (0, 4, 0):
            # msgpack only supports 'encoding' starting in 0.4.0.
            # Due to this, if we don't need it, don't pass it at all so
            # that under Python 2 we can still work with older versions
            # of msgpack.
            if salt.utils.msgpack.version >= (0, 5, 2):
                if encoding is None:
                    loads_kwargs["raw"] = True
                else:
                    loads_kwargs["raw"] = False
            else:
                loads_kwargs["encoding"] = encoding
            try:
                ret = salt.utils.msgpack.unpackb(msg, **loads_kwargs)
            except UnicodeDecodeError:
                # msg contains binary data
                loads_kwargs.pop("raw", None)
                loads_kwargs.pop("encoding", None)
                ret = salt.utils.msgpack.loads(msg, **loads_kwargs)
        else:
            ret = salt.utils.msgpack.loads(msg, **loads_kwargs)
        if encoding is None and not raw:
            ret = salt.transport.frame.decode_embedded_strs(ret)
    except Exception as exc:  # pylint: disable=broad-except
        log.critical(
            "Could not deserialize msgpack message. This often happens "
            "when trying to read a file not in binary mode. "
            "To see message payload, enable debug logging and retry. "
            "Exception: %s",
            exc,
        )
        log.debug("Msgpack deserialization failure on message: %s", msg)
        exc_msg = "Could not deserialize msgpack message. See log for more info."
        raise SaltDeserializationError(exc_msg) from exc
    finally:
        gc.enable()
    return ret


def dumps(msg, use_bin_type=False):
    """
    Run the correct dumps serialization format

    :param use_bin_type: Useful for Python 3 support. Tells msgpack to
                         differentiate between 'str' and 'bytes' types
                         by encoding them differently.
                         Since this changes the wire protocol, this
                         option should not be used outside of IPC.
    """

    def ext_type_encoder(obj):
        if isinstance(obj, int):
            # msgpack can't handle the very long Python longs for jids
            # Convert any very long longs to strings
            return str(obj)
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            # msgpack doesn't support datetime.datetime and datetime.date datatypes.
            # So here we have converted these types to custom datatype
            # This is msgpack Extended types numbered 78
            return salt.utils.msgpack.ExtType(
                78,
                salt.utils.stringutils.to_bytes(obj.strftime("%Y%m%dT%H:%M:%S.%f")),
            )
        elif isinstance(obj, _Constant):
            # Special case our constants.
            return salt.utils.msgpack.ExtType(
                79,
                salt.utils.msgpack.dumps((obj.name, obj.value), use_bin_type=True),
            )
        # The same for immutable types
        elif isinstance(obj, immutabletypes.ImmutableDict):
            return dict(obj)
        elif isinstance(obj, immutabletypes.ImmutableList):
            return list(obj)
        elif isinstance(obj, (set, immutabletypes.ImmutableSet)):
            # msgpack can't handle set so translate it to tuple
            return tuple(obj)
        elif isinstance(obj, CaseInsensitiveDict):
            return dict(obj)
        elif isinstance(obj, collections.abc.MutableMapping):
            return dict(obj)
        # Nothing known exceptions found. Let msgpack raise its own.
        return obj

    try:
        return salt.utils.msgpack.packb(
            msg, default=ext_type_encoder, use_bin_type=use_bin_type
        )
    except (OverflowError, salt.utils.msgpack.exceptions.PackValueError):
        # msgpack<=0.4.6 don't call ext encoder on very long integers raising the error instead.
        # Convert any very long longs to strings and call dumps again.
        def verylong_encoder(obj, context):
            # Make sure we catch recursion here.
            objid = id(obj)
            # This instance list needs to correspond to the types recursed
            # in the below if/elif chain. Also update
            # tests/unit/test_payload.py
            if objid in context and isinstance(obj, (dict, list, tuple)):
                return "<Recursion on {} with id={}>".format(
                    type(obj).__name__, id(obj)
                )
            context.add(objid)

            # The isinstance checks in this if/elif chain need to be
            # kept in sync with the above recursion check.
            if isinstance(obj, dict):
                for key, value in obj.copy().items():
                    obj[key] = verylong_encoder(value, context)
                return dict(obj)
            elif isinstance(obj, (list, tuple)):
                obj = list(obj)
                for idx, entry in enumerate(obj):
                    obj[idx] = verylong_encoder(entry, context)
                return obj
            # A value of an Integer object is limited from -(2^63) upto (2^64)-1 by MessagePack
            # spec. Here we care only of JIDs that are positive integers.
            if isinstance(obj, int) and obj >= pow(2, 64):
                return str(obj)
            else:
                return obj

        msg = verylong_encoder(msg, set())
        return salt.utils.msgpack.packb(
            msg, default=ext_type_encoder, use_bin_type=use_bin_type
        )


def load(fn_):
    """
    Run the correct serialization to load a file
    """
    data = fn_.read()
    fn_.close()
    if data:
        return loads(data, encoding="utf-8")


def dump(msg, fn_):
    """
    Serialize the correct data into the named file object
    """
    # When using Python 3, write files in such a way
    # that the 'bytes' and 'str' types are distinguishable
    # by using "use_bin_type=True".
    fn_.write(dumps(msg, use_bin_type=True))
    fn_.close()


class Serial:
    """
    Create a serialization object, this object manages all message
    serialization in Salt
    """

    def __init__(self, *args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "The `salt.payload.Serial` class has been deprecated, "
            "and is set to be removed in {version}. "
            "Please use `salt.payload.loads` and `salt.payload.dumps`.",
        )

    loads = staticmethod(loads)
    dumps = staticmethod(dumps)
    dump = staticmethod(dump)
    load = staticmethod(load)


class SREQ:
    """
    Create a generic interface to wrap salt zeromq req calls.
    """

    def __init__(self, master, id_="", serial="msgpack", linger=0, opts=None):
        self.master = master
        self.id_ = id_
        self.linger = linger
        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.opts = opts

    @property
    def socket(self):
        """
        Lazily create the socket.
        """
        if not hasattr(self, "_socket"):
            # create a new one
            self._socket = self.context.socket(zmq.REQ)
            if hasattr(zmq, "RECONNECT_IVL_MAX"):
                self._socket.setsockopt(zmq.RECONNECT_IVL_MAX, 5000)

            self._set_tcp_keepalive()
            if self.master.startswith("tcp://["):
                # Hint PF type if bracket enclosed IPv6 address
                if hasattr(zmq, "IPV6"):
                    self._socket.setsockopt(zmq.IPV6, 1)
                elif hasattr(zmq, "IPV4ONLY"):
                    self._socket.setsockopt(zmq.IPV4ONLY, 0)
            self._socket.linger = self.linger
            if self.id_:
                self._socket.setsockopt(zmq.IDENTITY, self.id_)
            self._socket.connect(self.master)
        return self._socket

    def _set_tcp_keepalive(self):
        if hasattr(zmq, "TCP_KEEPALIVE") and self.opts:
            if "tcp_keepalive" in self.opts:
                self._socket.setsockopt(zmq.TCP_KEEPALIVE, self.opts["tcp_keepalive"])
            if "tcp_keepalive_idle" in self.opts:
                self._socket.setsockopt(
                    zmq.TCP_KEEPALIVE_IDLE, self.opts["tcp_keepalive_idle"]
                )
            if "tcp_keepalive_cnt" in self.opts:
                self._socket.setsockopt(
                    zmq.TCP_KEEPALIVE_CNT, self.opts["tcp_keepalive_cnt"]
                )
            if "tcp_keepalive_intvl" in self.opts:
                self._socket.setsockopt(
                    zmq.TCP_KEEPALIVE_INTVL, self.opts["tcp_keepalive_intvl"]
                )

    def clear_socket(self):
        """
        delete socket if you have it
        """
        if hasattr(self, "_socket"):
            if isinstance(self.poller.sockets, dict):
                sockets = list(self.poller.sockets.keys())
                for socket in sockets:
                    log.trace("Unregistering socket: %s", socket)
                    self.poller.unregister(socket)
            else:
                for socket in self.poller.sockets:
                    log.trace("Unregistering socket: %s", socket)
                    self.poller.unregister(socket[0])
            del self._socket

    def send(self, enc, load, tries=1, timeout=60):
        """
        Takes two arguments, the encryption type and the base payload
        """
        payload = {"enc": enc}
        payload["load"] = load
        pkg = dumps(payload)
        self.socket.send(pkg)
        self.poller.register(self.socket, zmq.POLLIN)
        tried = 0
        while True:
            polled = self.poller.poll(timeout * 1000)
            tried += 1
            if polled:
                break
            if tries > 1:
                log.info(
                    "SaltReqTimeoutError: after %s seconds. (Try %s of %s)",
                    timeout,
                    tried,
                    tries,
                )
            if tried >= tries:
                self.clear_socket()
                raise SaltReqTimeoutError(
                    "SaltReqTimeoutError: after {} seconds, ran {} tries".format(
                        timeout * tried, tried
                    )
                )
        return loads(self.socket.recv())

    def send_auto(self, payload, tries=1, timeout=60):
        """
        Detect the encryption type based on the payload
        """
        enc = payload.get("enc", "clear")
        load = payload.get("load", {})
        return self.send(enc, load, tries, timeout)

    def destroy(self):
        if isinstance(self.poller.sockets, dict):
            sockets = list(self.poller.sockets.keys())
            for socket in sockets:
                if socket.closed is False:
                    socket.setsockopt(zmq.LINGER, 1)
                    socket.close()
                self.poller.unregister(socket)
        else:
            for socket in self.poller.sockets:
                if socket[0].closed is False:
                    socket[0].setsockopt(zmq.LINGER, 1)
                    socket[0].close()
                self.poller.unregister(socket[0])
        if self.socket.closed is False:
            self.socket.setsockopt(zmq.LINGER, 1)
            self.socket.close()
        if self.context.closed is False:
            self.context.term()

    # pylint: disable=W1701
    def __del__(self):
        self.destroy()

    # pylint: enable=W1701



"""
Unit tests for salt/modules/salt_version.py
"""

import pytest

import salt.modules.salt_version as salt_version
import salt.version
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


def test_mocked_objects():
    """
    Test that the mocked objects actually have what we expect.

    For example, earlier tests incorrectly mocked the
    salt.version.SaltStackVersion.LNAMES dict using upper-case indexes
    """
    assert isinstance(salt.version.SaltStackVersion.LNAMES, dict)
    sv = salt.version.SaltStackVersion(  # pylint: disable=no-value-for-parameter
        *salt.version.__version_info__
    )
    for k, v in salt.version.SaltStackVersion.LNAMES.items():
        assert k == k.lower()
        assert isinstance(v, tuple)
        if sv.new_version(major=v[0]):
            assert len(v) == 1
        else:
            assert len(v) == 2

    sv = sv.__str__()
    assert isinstance(sv, str)

    with patch("salt.version.SaltStackVersion.LNAMES", {"neon": (2019, 8)}):
        sv = salt.version.SaltStackVersion.from_name("Neon")
        assert sv.string == "2019.8.0"


def test_get_release_number_no_codename():
    """
    Test that None is returned when the codename isn't found.
    """
    assert salt_version.get_release_number("foo") is None


def test_get_release_number_unassigned():
    """
    Test that a string is returned when a version is found, but unassigned.
    """
    with patch("salt.version.SaltStackVersion.LNAMES", {"foo": (12345, 0)}):
        mock_str = "No version assigned."
        assert salt_version.get_release_number("foo") == mock_str


def test_get_release_number_success():
    """
    Test that a version is returned for a released codename
    """
    assert salt_version.get_release_number("Oxygen") == "2018.3"


def test_get_release_number_success_new_version():
    """
    Test that a version is returned for new versioning (3000)
    """
    assert salt_version.get_release_number("Neon") == "3000"


def test_get_release_number_success_new_version_with_dot():
    """
    Test that a version is returned for new versioning (3006)
    """
    assert salt_version.get_release_number("Sulfur") == "3006"


def test_equal_success():
    """
    Test that the current version is equal to the codename
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="1900.5.0")):
        with patch("salt.version.SaltStackVersion.LNAMES", {"foo": (1900, 5)}):
            assert salt_version.equal("foo") is True


def test_equal_success_new_version():
    """
    Test that the current version is equal to the codename
    while using the new versioning
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="3000.1")):
        with patch("salt.version.SaltStackVersion.LNAMES", {"foo": (3000,)}):
            assert salt_version.equal("foo") is True


def test_equal_success_new_version_with_dot():
    """
    Test that the current version is equal to the codename
    while using the new versioning
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="3006.1")):
        with patch("salt.version.SaltStackVersion.LNAMES", {"foo": (3006,)}):
            assert salt_version.equal("foo") is True


def test_equal_older_codename():
    """
    Test that when an older codename is passed in, the function returns False.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.version.SaltStackVersion.LNAMES",
            {"oxygen": (2018, 3), "nitrogen": (2017, 7)},
        ):
            assert salt_version.equal("Nitrogen") is False


def test_equal_older_codename_new_version():
    """
    Test that when an older codename is passed in, the function returns False.
    while also testing with the new versioning.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.version.SaltStackVersion.LNAMES",
            {"neon": (3000), "nitrogen": (2017, 7)},
        ):
            assert salt_version.equal("Nitrogen") is False


def test_equal_newer_codename():
    """
    Test that when a newer codename is passed in, the function returns False
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.version.SaltStackVersion.LNAMES",
            {"fluorine": (salt.version.MAX_SIZE - 100, 0)},
        ):
            assert salt_version.equal("Fluorine") is False


def test_greater_than_success():
    """
    Test that the current version is newer than the codename
    """
    with patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2017.7")
    ):
        with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
            assert salt_version.greater_than("Nitrogen") is True


def test_greater_than_success_new_version():
    """
    Test that the current version is newer than the codename
    """
    with patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="2017.7")
    ):
        with patch("salt.version.SaltStackVersion", MagicMock(return_value="3000")):
            assert salt_version.greater_than("Nitrogen") is True


def test_greater_than_success_new_version_with_dot():
    """
    Test that the current version is newer than the codename
    """
    with patch(
        "salt.modules.salt_version.get_release_number", MagicMock(return_value="3000")
    ):
        with patch("salt.version.SaltStackVersion", MagicMock(return_value="3006.0")):
            assert salt_version.greater_than("Neon") is True


def test_greater_than_with_equal_codename():
    """
    Test that when an equal codename is passed in, the function returns False.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch("salt.version.SaltStackVersion.LNAMES", {"oxygen": (2018, 3)}):
            assert salt_version.greater_than("Oxygen") is False


def test_greater_than_with_newer_codename():
    """
    Test that when a newer codename is passed in, the function returns False.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.version.SaltStackVersion.LNAMES",
            {"fluorine": (2019, 2), "oxygen": (2018, 3)},
        ):
            assert salt_version.greater_than("Fluorine") is False


def test_greater_than_unassigned():
    """
    Test that the unassigned codename is greater than the current version
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="No version assigned."),
        ):
            assert salt_version.greater_than("Fluorine") is False


def test_less_than_success():
    """
    Test that when a newer codename is passed in, the function returns True.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="2019.2"),
        ):
            assert salt_version.less_than("Fluorine") is True


def test_less_than_success_new_version():
    """
    Test that when a newer codename is passed in, the function returns True
    using new version
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="3000"),
        ):
            assert salt_version.less_than("Fluorine") is True


def test_less_than_success_new_version_with_dot():
    """
    Test that when a newer codename is passed in, the function returns True
    using new version
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="3006"),
        ):
            assert salt_version.less_than("Fluorine") is True


def test_less_than_do_not_crash_when_input_is_a_number():
    """
    Test that less_than do not crash when unexpected inputs
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with pytest.raises(CommandExecutionError):
            salt_version.less_than(1234)


def test_less_than_with_equal_codename():
    """
    Test that when an equal codename is passed in, the function returns False.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch("salt.version.SaltStackVersion.LNAMES", {"oxygen": (2018, 3)}):
            assert salt_version.less_than("Oxygen") is False


def test_less_than_with_older_codename():
    """
    Test that the current version is less than the codename.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="2017.7"),
        ):
            assert salt_version.less_than("Nitrogen") is False


def test_less_than_with_unassigned_codename():
    """
    Test that when an unassigned codename greater than the current version.
    """
    with patch("salt.version.SaltStackVersion", MagicMock(return_value="2018.3.2")):
        with patch(
            "salt.modules.salt_version.get_release_number",
            MagicMock(return_value="No version assigned."),
        ):
            assert salt_version.less_than("Fluorine") is True


def test_check_release_cmp_no_codename():
    """
    Test that None is returned when the codename isn't found.
    """
    assert salt_version._check_release_cmp("foo") is None


def test_check_release_cmp_success():
    """
    Test that an int is returned from the version compare
    """
    assert isinstance(salt_version._check_release_cmp("Oxygen"), int)



import datetime

import pytest

import salt.modules.win_event as win_event

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="function")
def application_events():
    # This deletes the contents of the Application event log
    win_event.clear("Application")
    win_event.add("Application", 2011, event_type="Information")
    win_event.add("Application", 2011, event_type="Information")
    win_event.add("Application", 2011, event_type="Information")
    win_event.add("Application", 2011, event_type="Information")
    win_event.add("Application", 2020, event_type="Warning")
    win_event.add("Application", 2020, event_type="Warning")
    yield
    # This deletes the contents of the Application event log
    win_event.clear("Application")


def test__to_bytes_utf8():
    data = {"key1": "item1", "key2": [1, 2, "item2"], "key3": 45, 45: str}

    new_data = win_event._to_bytes(data, "utf-8", False)

    assert "key1" in new_data

    assert new_data["key1"] == b"item1"
    assert new_data["key2"][2] == b"item2"


def test__to_bytes_cp1252():
    data = {"key1": "item1", "key2": [1, 2, "item2"], "key3": 45, 45: str}

    new_data = win_event._to_bytes(data, "CP1252", True)

    assert b"key1" in new_data
    assert b"key2" in new_data
    assert b"key3" in new_data

    assert new_data["key1".encode("CP1252")] == "item1".encode("CP1252")
    assert new_data["key2".encode("CP1252")][2] == "item2".encode("CP1252")


def test__raw_time():
    raw_time = win_event._raw_time(datetime.datetime(2019, 7, 2, 10, 8, 19))
    assert raw_time == (2019, 7, 2, 10, 8, 19)


@pytest.mark.destructive_test
def test_count(application_events):
    """
    Test win_event.count
    """
    ret = win_event.count("Application")
    assert ret == 6


@pytest.mark.destructive_test
def test_get(application_events):
    ret = win_event.get("Application")
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_query(application_events):
    ret = win_event.query("Application")
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_query_records(application_events):
    ret = win_event.query("Application", records=3)
    for item in ret:
        assert isinstance(item, dict)
    assert len(ret) == 3


@pytest.mark.destructive_test
def test_query_raw(application_events):
    ret = win_event.query("Application", raw=True)
    for item in ret:
        assert isinstance(item, str)
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_query_level(application_events):
    ret = win_event.query("Application", "*[System[(Level=3)]]")
    assert len(ret) == 2


@pytest.mark.destructive_test
def test_query_level_eventid(application_events):
    ret = win_event.query(
        "Application", "*[System[(Level=4 or Level=0) and (EventID=2011)]]"
    )
    assert len(ret) == 4


@pytest.mark.destructive_test
def test_query_last_hour(application_events):
    ret = win_event.query(
        "Application", "*[System[TimeCreated[timediff(@SystemTime) <= 3600000]]]"
    )
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_get_filtered(application_events):
    ret = win_event.get_filtered("Application")
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_get_filtered_event_id(application_events):
    ret = win_event.get_filtered("Application", eventID=2011)
    assert len(ret) == 4


@pytest.mark.destructive_test
def test_get_filtered_event_type(application_events):
    ret = win_event.get_filtered("Application", eventType=2)
    assert len(ret) == 2


@pytest.mark.destructive_test
def test_get_filtered_year(application_events):
    year = datetime.datetime.now().year
    ret = win_event.get_filtered("Application", year=year)
    assert len(ret) == 6


@pytest.mark.destructive_test
def test_get_filtered_year_none(application_events):
    year = 1999
    ret = win_event.get_filtered("Application", year=year)
    assert len(ret) == 0


@pytest.mark.destructive_test
def test_clear(application_events):
    assert win_event.count("Application") == 6
    win_event.clear("Application")
    assert win_event.count("Application") == 0


@pytest.mark.destructive_test
def test_clear_backup(application_events, tmp_path):
    assert win_event.count("Application") == 6
    backup_log = tmp_path / "test.bak"
    assert not backup_log.exists()
    win_event.clear("Application", str(backup_log))
    assert backup_log.exists()
    assert win_event.count("Application") == 0



"""
Integration tests for mac_timezone

If using parallels, make sure Time sync is turned off. Otherwise, parallels will
keep changing your date/time settings while the tests are running. To turn off
Time sync do the following:
    - Go to actions -> configure
    - Select options at the top and 'More Options' on the left
    - Set time to 'Do not sync'
"""

import datetime

import pytest

from tests.support.case import ModuleCase


@pytest.mark.flaky(max_runs=4)
@pytest.mark.skip_unless_on_darwin
@pytest.mark.skip_if_binaries_missing("systemsetup")
@pytest.mark.skip_if_not_root
@pytest.mark.slow_test
class MacTimezoneModuleTest(ModuleCase):
    """
    Validate the mac_timezone module
    """

    USE_NETWORK_TIME = False
    TIME_SERVER = "time.apple.com"
    TIME_ZONE = ""
    CURRENT_DATE = ""
    CURRENT_TIME = ""

    def setUp(self):
        """
        Get current settings
        """
        self.USE_NETWORK_TIME = self.run_function("timezone.get_using_network_time")
        self.TIME_SERVER = self.run_function("timezone.get_time_server")
        self.TIME_ZONE = self.run_function("timezone.get_zone")
        self.CURRENT_DATE = self.run_function("timezone.get_date")
        self.CURRENT_TIME = self.run_function("timezone.get_time")

        self.run_function("timezone.set_using_network_time", [False])
        self.run_function("timezone.set_zone", ["America/Denver"])

    def tearDown(self):
        """
        Reset to original settings
        """
        self.run_function("timezone.set_time_server", [self.TIME_SERVER])
        self.run_function("timezone.set_using_network_time", [self.USE_NETWORK_TIME])
        self.run_function("timezone.set_zone", [self.TIME_ZONE])
        if not self.USE_NETWORK_TIME:
            self.run_function("timezone.set_date", [self.CURRENT_DATE])
            self.run_function("timezone.set_time", [self.CURRENT_TIME])

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_set_date(self):
        """
        Test timezone.get_date
        Test timezone.set_date
        """
        # Correct Functionality
        self.assertTrue(self.run_function("timezone.set_date", ["2/20/2011"]))
        self.assertEqual(self.run_function("timezone.get_date"), "2/20/2011")

        # Test bad date format
        self.assertEqual(
            self.run_function("timezone.set_date", ["13/12/2014"]),
            "ERROR executing 'timezone.set_date': Invalid Date/Time Format: 13/12/2014",
        )

    @pytest.mark.slow_test
    def test_get_time(self):
        """
        Test timezone.get_time
        """
        text_time = self.run_function("timezone.get_time")
        self.assertNotEqual(text_time, "Invalid Timestamp")
        obj_date = datetime.datetime.strptime(text_time, "%H:%M:%S")
        self.assertIsInstance(obj_date, datetime.date)

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_set_time(self):
        """
        Test timezone.set_time
        """
        # Correct Functionality
        self.assertTrue(self.run_function("timezone.set_time", ["3:14"]))

        # Test bad time format
        self.assertEqual(
            self.run_function("timezone.set_time", ["3:71"]),
            "ERROR executing 'timezone.set_time': Invalid Date/Time Format: 3:71",
        )

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_set_zone(self):
        """
        Test timezone.get_zone
        Test timezone.set_zone
        """
        # Correct Functionality
        self.assertTrue(self.run_function("timezone.set_zone", ["Pacific/Wake"]))
        self.assertEqual(self.run_function("timezone.get_zone"), "Pacific/Wake")

        # Test bad time zone
        self.assertEqual(
            self.run_function("timezone.set_zone", ["spongebob"]),
            "ERROR executing 'timezone.set_zone': Invalid Timezone: spongebob",
        )

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_offset(self):
        """
        Test timezone.get_offset
        """
        self.assertTrue(self.run_function("timezone.set_zone", ["Pacific/Wake"]))
        self.assertIsInstance(self.run_function("timezone.get_offset"), (str,))
        self.assertEqual(self.run_function("timezone.get_offset"), "+1200")

        self.assertTrue(self.run_function("timezone.set_zone", ["America/Los_Angeles"]))
        self.assertIsInstance(self.run_function("timezone.get_offset"), (str,))
        self.assertEqual(self.run_function("timezone.get_offset"), "-0700")

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_set_zonecode(self):
        """
        Test timezone.get_zonecode
        Test timezone.set_zonecode
        """
        self.assertTrue(self.run_function("timezone.set_zone", ["America/Los_Angeles"]))
        self.assertIsInstance(self.run_function("timezone.get_zonecode"), (str,))
        self.assertEqual(self.run_function("timezone.get_zonecode"), "PDT")

        self.assertTrue(self.run_function("timezone.set_zone", ["Pacific/Wake"]))
        self.assertIsInstance(self.run_function("timezone.get_zonecode"), (str,))
        self.assertEqual(self.run_function("timezone.get_zonecode"), "WAKT")

    @pytest.mark.slow_test
    def test_list_zones(self):
        """
        Test timezone.list_zones
        """
        zones = self.run_function("timezone.list_zones")
        self.assertIsInstance(self.run_function("timezone.list_zones"), list)
        self.assertIn("America/Denver", self.run_function("timezone.list_zones"))
        self.assertIn("America/Los_Angeles", self.run_function("timezone.list_zones"))

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_zone_compare(self):
        """
        Test timezone.zone_compare
        """
        self.assertTrue(self.run_function("timezone.set_zone", ["America/Denver"]))
        self.assertTrue(self.run_function("timezone.zone_compare", ["America/Denver"]))
        self.assertFalse(self.run_function("timezone.zone_compare", ["Pacific/Wake"]))

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_set_using_network_time(self):
        """
        Test timezone.get_using_network_time
        Test timezone.set_using_network_time
        """
        self.assertTrue(self.run_function("timezone.set_using_network_time", [True]))
        self.assertTrue(self.run_function("timezone.get_using_network_time"))

        self.assertTrue(self.run_function("timezone.set_using_network_time", [False]))
        self.assertFalse(self.run_function("timezone.get_using_network_time"))

    @pytest.mark.skip(
        reason="Skip until we can figure out why modifying the system clock causes ZMQ errors",
    )
    @pytest.mark.destructive_test
    def test_get_set_time_server(self):
        """
        Test timezone.get_time_server
        Test timezone.set_time_server
        """
        self.assertTrue(
            self.run_function("timezone.set_time_server", ["spongebob.com"])
        )
        self.assertEqual(self.run_function("timezone.get_time_server"), "spongebob.com")



"""
Execute an unmodified puppet_node_classifier and read the output as YAML. The YAML data is then directly overlaid onto the minion's Pillar data.
"""

import logging

import salt.utils.yaml

# Don't "fix" the above docstring to put it on two lines, as the sphinx
# autosummary pulls only the first line for its description.


# Set up logging
log = logging.getLogger(__name__)


def ext_pillar(minion_id, pillar, command):  # pylint: disable=W0613
    """
    Execute an unmodified puppet_node_classifier and read the output as YAML
    """
    try:
        data = salt.utils.yaml.safe_load(
            __salt__["cmd.run"]("{} {}".format(command, minion_id))
        )
        return data["parameters"]
    except Exception:  # pylint: disable=broad-except
        log.critical("YAML data from %s failed to parse", command)
        return {}



"""
macOS implementations of various commands in the "desktop" interface
"""

import salt.utils.platform
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "desktop"


def __virtual__():
    """
    Only load on Mac systems
    """
    if salt.utils.platform.is_darwin():
        return __virtualname__
    return False, "Cannot load macOS desktop module: This is not a macOS host."


def get_output_volume():
    """
    Get the output volume (range 0 to 100)

    CLI Example:

    .. code-block:: bash

        salt '*' desktop.get_output_volume
    """
    cmd = 'osascript -e "get output volume of (get volume settings)"'
    call = __salt__["cmd.run_all"](cmd, output_loglevel="debug", python_shell=False)
    _check_cmd(call)

    return call.get("stdout")


def set_output_volume(volume):
    """
    Set the volume of sound.

    volume
        The level of volume. Can range from 0 to 100.

    CLI Example:

    .. code-block:: bash

        salt '*' desktop.set_output_volume <volume>
    """
    cmd = 'osascript -e "set volume output volume {}"'.format(volume)
    call = __salt__["cmd.run_all"](cmd, output_loglevel="debug", python_shell=False)
    _check_cmd(call)

    return get_output_volume()


def screensaver():
    """
    Launch the screensaver.

    CLI Example:

    .. code-block:: bash

        salt '*' desktop.screensaver
    """
    cmd = "open /System/Library/Frameworks/ScreenSaver.framework/Versions/A/Resources/ScreenSaverEngine.app"
    call = __salt__["cmd.run_all"](cmd, output_loglevel="debug", python_shell=False)
    _check_cmd(call)

    return True


def lock():
    """
    Lock the desktop session

    CLI Example:

    .. code-block:: bash

        salt '*' desktop.lock
    """
    cmd = (
        "/System/Library/CoreServices/Menu\\"
        " Extras/User.menu/Contents/Resources/CGSession -suspend"
    )
    call = __salt__["cmd.run_all"](cmd, output_loglevel="debug", python_shell=False)
    _check_cmd(call)

    return True


def say(*words):
    """
    Say some words.

    words
        The words to execute the say command with.

    CLI Example:

    .. code-block:: bash

        salt '*' desktop.say <word0> <word1> ... <wordN>
    """
    cmd = "say {}".format(" ".join(words))
    call = __salt__["cmd.run_all"](cmd, output_loglevel="debug", python_shell=False)
    _check_cmd(call)

    return True


def _check_cmd(call):
    """
    Check the output of the cmd.run_all function call.
    """
    if call["retcode"] != 0:
        comment = ""
        std_err = call.get("stderr")
        std_out = call.get("stdout")
        if std_err:
            comment += std_err
        if std_out:
            comment += std_out

        raise CommandExecutionError("Error running command: {}".format(comment))

    return call



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""
import pytest

import salt.modules.libcloud_loadbalancer as libcloud_loadbalancer
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    from libcloud.loadbalancer.base import Algorithm, BaseDriver, LoadBalancer, Member

    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False

if HAS_LIBCLOUD:

    class MockLBDriver(BaseDriver):
        def __init__(self):  # pylint: disable=W0231
            self._TEST_BALANCER = LoadBalancer(
                id="test_id",
                name="test_balancer",
                state=0,  # RUNNING
                ip="1.2.3.4",
                port=80,
                driver=self,
                extra={},
            )
            self._TEST_MEMBER = Member(
                id="member_id",
                ip="12.3.4.5",
                port=443,
                balancer=self._TEST_BALANCER,
                extra=None,
            )

        def get_balancer(self, balancer_id):
            assert balancer_id == "test_id"
            return self._TEST_BALANCER

        def list_balancers(self):
            return [self._TEST_BALANCER]

        def list_protocols(self):
            return ["http", "https"]

        def create_balancer(self, name, port, protocol, algorithm, members):
            assert name == "new_test_balancer"
            assert port == 80
            assert protocol == "http"
            assert isinstance(algorithm, (Algorithm, int))
            assert isinstance(members, list)
            return self._TEST_BALANCER

        def destroy_balancer(self, balancer):
            assert balancer == self._TEST_BALANCER
            return True

        def balancer_attach_member(self, balancer, member):
            assert isinstance(balancer, LoadBalancer)
            assert isinstance(member, Member)
            assert member.id is None
            assert balancer.id == "test_id"
            return self._TEST_MEMBER

        def balancer_detach_member(self, balancer, member):
            assert isinstance(balancer, LoadBalancer)
            assert isinstance(member, Member)
            assert member.id == "member_id"
            assert balancer.id == "test_id"
            return True

        def balancer_list_members(self, balancer):
            assert isinstance(balancer, LoadBalancer)
            assert balancer.id == "test_id"
            return [self._TEST_MEMBER]

else:
    MockLBDriver = object


def get_mock_driver():
    return MockLBDriver()


@pytest.mark.skipif(not HAS_LIBCLOUD, reason="No libcloud package")
@patch(
    "salt.modules.libcloud_loadbalancer._get_driver",
    MagicMock(return_value=MockLBDriver()),
)
class LibcloudLoadBalancerModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                )
            }
        }
        if libcloud_loadbalancer.HAS_LIBCLOUD is False:
            module_globals["sys.modules"] = {"libcloud": MagicMock()}

        return {libcloud_loadbalancer: module_globals}

    def test_module_creation(self):
        client = libcloud_loadbalancer._get_driver("test")
        self.assertFalse(client is None)

    def _validate_balancer(self, balancer):
        self.assertEqual(balancer["name"], "test_balancer")

    def _validate_member(self, member):
        self.assertEqual(member["id"], "member_id")
        self.assertEqual(member["ip"], "12.3.4.5")

    def test_list_balancers(self):
        balancers = libcloud_loadbalancer.list_balancers("test")
        self.assertEqual(len(balancers), 1)
        self._validate_balancer(balancers[0])

    def test_list_protocols(self):
        protocols = libcloud_loadbalancer.list_protocols("test")
        self.assertEqual(len(protocols), 2)
        self.assertTrue("http" in protocols)

    def test_create_balancer(self):
        balancer = libcloud_loadbalancer.create_balancer(
            "new_test_balancer", 80, "http", "test"
        )
        self._validate_balancer(balancer)

    def test_create_balancer_custom_algorithm(self):
        balancer = libcloud_loadbalancer.create_balancer(
            "new_test_balancer", 80, "http", "test", algorithm="LEAST_CONNECTIONS"
        )
        self._validate_balancer(balancer)

    def test_destroy_balancer(self):
        result = libcloud_loadbalancer.destroy_balancer("test_id", "test")
        self.assertTrue(result)

    def test_get_balancer_by_name(self):
        balancer = libcloud_loadbalancer.get_balancer_by_name("test_balancer", "test")
        self._validate_balancer(balancer)

    def test_get_balancer(self):
        balancer = libcloud_loadbalancer.get_balancer("test_id", "test")
        self._validate_balancer(balancer)

    def test_balancer_attach_member(self):
        member = libcloud_loadbalancer.balancer_attach_member(
            "test_id", "12.3.4.5", 443, "test"
        )
        self._validate_member(member)

    def test_balancer_detach_member(self):
        result = libcloud_loadbalancer.balancer_detach_member(
            "test_id", "member_id", "test"
        )
        self.assertTrue(result)

    def test_list_balancer_members(self):
        members = libcloud_loadbalancer.list_balancer_members("test_id", "test")
        self._validate_member(members[0])



"""
States for managing zfs datasets

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs, salt.modules.zfs
:platform:      smartos, illumos, solaris, freebsd, linux

.. versionadded:: 2016.3.0
.. versionchanged:: 2018.3.1
  Big refactor to remove duplicate code, better type conversions and improved
  consistency in output.

.. code-block:: yaml

    test/shares/yuki:
      zfs.filesystem_present:
        - create_parent: true
        - properties:
            quota: 16G

    test/iscsi/haruhi:
      zfs.volume_present:
        - create_parent: true
        - volume_size: 16M
        - sparse: true
        - properties:
            readonly: on

    test/shares/yuki@frozen:
      zfs.snapshot_present

    moka_origin:
      zfs.hold_present:
        - snapshot: test/shares/yuki@frozen

    test/shares/moka:
      zfs.filesystem_present:
        - cloned_from: test/shares/yuki@frozen

    test/shares/moka@tsukune:
      zfs.snapshot_absent

"""

import logging
from datetime import datetime

from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

# Define the state's virtual name
__virtualname__ = "zfs"

# Compare modifiers for zfs.schedule_snapshot
comp_hour = {"minute": 0}
comp_day = {"minute": 0, "hour": 0}
comp_month = {"minute": 0, "hour": 0, "day": 1}
comp_year = {"minute": 0, "hour": 0, "day": 1, "month": 1}


def __virtual__():
    """
    Provides zfs state
    """
    if not __grains__.get("zfs_support"):
        return False, "The zfs state cannot be loaded: zfs not supported"
    return __virtualname__


def _absent(name, dataset_type, force=False, recursive=False):
    """
    internal shared function for *_absent

    name : string
        name of dataset
    dataset_type : string [filesystem, volume, snapshot, or bookmark]
        type of dataset to remove
    force : boolean
        try harder to destroy the dataset
    recursive : boolean
        also destroy all the child datasets

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## log configuration
    dataset_type = dataset_type.lower()
    log.debug("zfs.%s_absent::%s::config::force = %s", dataset_type, name, force)
    log.debug(
        "zfs.%s_absent::%s::config::recursive = %s", dataset_type, name, recursive
    )

    ## destroy dataset if needed
    if __salt__["zfs.exists"](name, **{"type": dataset_type}):
        ## NOTE: dataset found with the name and dataset_type
        if not __opts__["test"]:
            mod_res = __salt__["zfs.destroy"](
                name, **{"force": force, "recursive": recursive}
            )
        else:
            mod_res = OrderedDict([("destroyed", True)])

        ret["result"] = mod_res["destroyed"]
        if ret["result"]:
            ret["changes"][name] = "destroyed"
            ret["comment"] = "{} {} was destroyed".format(
                dataset_type,
                name,
            )
        else:
            ret["comment"] = "failed to destroy {} {}".format(
                dataset_type,
                name,
            )
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]
    else:
        ## NOTE: no dataset found with name of the dataset_type
        ret["comment"] = "{} {} is absent".format(dataset_type, name)

    return ret


def filesystem_absent(name, force=False, recursive=False):
    """
    ensure filesystem is absent on the system

    name : string
        name of filesystem
    force : boolean
        try harder to destroy the dataset (zfs destroy -f)
    recursive : boolean
        also destroy all the child datasets (zfs destroy -r)

    .. warning::

        If a volume with ``name`` exists, this state will succeed without
        destroying the volume specified by ``name``. This module is dataset type sensitive.

    """
    if not __utils__["zfs.is_dataset"](name):
        ret = {
            "name": name,
            "changes": {},
            "result": False,
            "comment": "invalid dataset name: {}".format(name),
        }
    else:
        ret = _absent(name, "filesystem", force, recursive)
    return ret


def volume_absent(name, force=False, recursive=False):
    """
    ensure volume is absent on the system

    name : string
        name of volume
    force : boolean
        try harder to destroy the dataset (zfs destroy -f)
    recursive : boolean
        also destroy all the child datasets (zfs destroy -r)

    .. warning::

        If a filesystem with ``name`` exists, this state will succeed without
        destroying the filesystem specified by ``name``. This module is dataset type sensitive.

    """
    if not __utils__["zfs.is_dataset"](name):
        ret = {
            "name": name,
            "changes": {},
            "result": False,
            "comment": "invalid dataset name: {}".format(name),
        }
    else:
        ret = _absent(name, "volume", force, recursive)
    return ret


def snapshot_absent(name, force=False, recursive=False):
    """
    ensure snapshot is absent on the system

    name : string
        name of snapshot
    force : boolean
        try harder to destroy the dataset (zfs destroy -f)
    recursive : boolean
        also destroy all the child datasets (zfs destroy -r)

    """
    if not __utils__["zfs.is_snapshot"](name):
        ret = {
            "name": name,
            "changes": {},
            "result": False,
            "comment": "invalid snapshot name: {}".format(name),
        }
    else:
        ret = _absent(name, "snapshot", force, recursive)
    return ret


def bookmark_absent(name, force=False, recursive=False):
    """
    ensure bookmark is absent on the system

    name : string
        name of snapshot
    force : boolean
        try harder to destroy the dataset (zfs destroy -f)
    recursive : boolean
        also destroy all the child datasets (zfs destroy -r)

    """
    if not __utils__["zfs.is_bookmark"](name):
        ret = {
            "name": name,
            "changes": {},
            "result": False,
            "comment": "invalid bookmark name: {}".format(name),
        }
    else:
        ret = _absent(name, "bookmark", force, recursive)
    return ret


def hold_absent(name, snapshot, recursive=False):
    """
    ensure hold is absent on the system

    name : string
        name of hold
    snapshot : string
        name of snapshot
    recursive : boolean
        recursively releases a hold with the given tag on the snapshots of all descendent file systems.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## log configuration
    log.debug("zfs.hold_absent::%s::config::snapshot = %s", name, snapshot)
    log.debug("zfs.hold_absent::%s::config::recursive = %s", name, recursive)

    ## check we have a snapshot/tag name
    if not __utils__["zfs.is_snapshot"](snapshot):
        ret["result"] = False
        ret["comment"] = "invalid snapshot name: {}".format(snapshot)
        return ret

    if (
        __utils__["zfs.is_snapshot"](name)
        or __utils__["zfs.is_bookmark"](name)
        or name == "error"
    ):
        ret["result"] = False
        ret["comment"] = "invalid tag name: {}".format(name)
        return ret

    ## release hold if required
    holds = __salt__["zfs.holds"](snapshot)
    if name in holds:
        ## NOTE: hold found for snapshot, release it
        if not __opts__["test"]:
            mod_res = __salt__["zfs.release"](
                name, snapshot, **{"recursive": recursive}
            )
        else:
            mod_res = OrderedDict([("released", True)])

        ret["result"] = mod_res["released"]
        if ret["result"]:
            ret["changes"] = {snapshot: {name: "released"}}
            ret["comment"] = "hold {} released".format(
                name,
            )
        else:
            ret["comment"] = "failed to release hold {}".format(
                name,
            )
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]
    elif "error" in holds:
        ## NOTE: we have an error
        ret["result"] = False
        ret["comment"] = holds["error"]
    else:
        ## NOTE: no hold found with name for snapshot
        ret["comment"] = "hold {} is absent".format(
            name,
        )

    return ret


def hold_present(name, snapshot, recursive=False):
    """
    ensure hold is present on the system

    name : string
        name of holdt
    snapshot : string
        name of snapshot
    recursive : boolean
        recursively add hold with the given tag on the snapshots of all descendent file systems.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## log configuration
    log.debug("zfs.hold_present::%s::config::snapshot = %s", name, snapshot)
    log.debug("zfs.hold_present::%s::config::recursive = %s", name, recursive)

    ## check we have a snapshot/tag name
    if not __utils__["zfs.is_snapshot"](snapshot):
        ret["result"] = False
        ret["comment"] = "invalid snapshot name: {}".format(snapshot)
        return ret

    if (
        __utils__["zfs.is_snapshot"](name)
        or __utils__["zfs.is_bookmark"](name)
        or name == "error"
    ):
        ret["result"] = False
        ret["comment"] = "invalid tag name: {}".format(name)
        return ret

    ## place hold if required
    holds = __salt__["zfs.holds"](snapshot)
    if name in holds:
        ## NOTE: hold with name already exists for snapshot
        ret["comment"] = "hold {} is present for {}".format(
            name,
            snapshot,
        )
    else:
        ## NOTE: no hold found with name for snapshot
        if not __opts__["test"]:
            mod_res = __salt__["zfs.hold"](name, snapshot, **{"recursive": recursive})
        else:
            mod_res = OrderedDict([("held", True)])

        ret["result"] = mod_res["held"]
        if ret["result"]:
            ret["changes"] = OrderedDict([(snapshot, OrderedDict([(name, "held")]))])
            ret["comment"] = "hold {} added to {}".format(name, snapshot)
        else:
            ret["comment"] = "failed to add hold {} to {}".format(name, snapshot)
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]

    return ret


def _dataset_present(
    dataset_type,
    name,
    properties,
    volume_size=None,
    sparse=False,
    create_parent=False,
    cloned_from=None,
):
    """
    internal handler for filesystem_present/volume_present

    dataset_type : string
        volume or filesystem
    name : string
        name of volume
    volume_size : string
        size of volume
    sparse : boolean
        create sparse volume
    create_parent : boolean
        creates all the non-existing parent datasets.
        any property specified on the command line using the -o option is ignored.
    cloned_from : string
        name of snapshot to clone
    properties : dict
        additional zfs properties (-o)

    .. note::
        ``cloned_from`` is only use if the volume does not exist yet,
        when ``cloned_from`` is set after the volume exists it will be ignored.

    .. note::
        Properties do not get cloned, if you specify the properties in the state file
        they will be applied on a subsequent run.

        ``volume_size`` is considered a property, so the volume's size will be
        corrected when the properties get updated if it differs from the
        original volume.

        The sparse parameter is ignored when using ``cloned_from``.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## fallback dataset_type to filesystem if out of range
    if dataset_type not in ["filesystem", "volume"]:
        dataset_type = "filesystem"

    ## ensure properties are zfs values
    if properties is None:
        properties = {}
    properties = __utils__["zfs.from_auto_dict"](properties)
    if volume_size:
        ## NOTE: add volsize to properties
        volume_size = __utils__["zfs.from_size"](volume_size)
        properties.update({"volsize": volume_size})
    # The sorted isn't necessary for proper behavior, but it helps for the unit
    # tests.
    propnames = ",".join(sorted(properties.keys()))

    ## log configuration
    log.debug(
        "zfs.%s_present::%s::config::volume_size = %s", dataset_type, name, volume_size
    )
    log.debug("zfs.%s_present::%s::config::sparse = %s", dataset_type, name, sparse)
    log.debug(
        "zfs.%s_present::%s::config::create_parent = %s",
        dataset_type,
        name,
        create_parent,
    )
    log.debug(
        "zfs.%s_present::%s::config::cloned_from = %s", dataset_type, name, cloned_from
    )
    log.debug(
        "zfs.%s_present::%s::config::properties = %s", dataset_type, name, properties
    )

    ## check we have valid filesystem name/volume name/clone snapshot
    if not __utils__["zfs.is_dataset"](name):
        ret["result"] = False
        ret["comment"] = "invalid dataset name: {}".format(name)
        return ret

    if cloned_from and not __utils__["zfs.is_snapshot"](cloned_from):
        ret["result"] = False
        ret["comment"] = "{} is not a snapshot".format(cloned_from)
        return ret

    ## ensure dataset is in correct state
    ## NOTE: update the dataset
    exists = __salt__["zfs.exists"](name, **{"type": dataset_type})
    if exists and len(properties) == 0:
        ret["comment"] = "{} {} is uptodate".format(dataset_type, name)
    elif exists and len(properties) > 0:
        ## NOTE: fetch current volume properties
        properties_current = __salt__["zfs.get"](
            name,
            properties=propnames,
            type=dataset_type,
            fields="value",
            depth=0,
            parsable=True,
        ).get(name, OrderedDict())

        ## NOTE: build list of properties to update
        properties_update = []
        for prop in properties:
            ## NOTE: skip unexisting properties
            if prop not in properties_current:
                log.warning(
                    "zfs.%s_present::%s::update - unknown property: %s",
                    dataset_type,
                    name,
                    prop,
                )
                continue

            ## NOTE: compare current and wanted value
            if properties_current[prop]["value"] != properties[prop]:
                properties_update.append(prop)

        ## NOTE: update pool properties
        for prop in properties_update:
            if not __opts__["test"]:
                mod_res = __salt__["zfs.set"](name, **{prop: properties[prop]})
            else:
                mod_res = OrderedDict([("set", True)])

            if mod_res["set"]:
                if name not in ret["changes"]:
                    ret["changes"][name] = {}
                ret["changes"][name][prop] = properties[prop]
            else:
                ret["result"] = False
                if ret["comment"] == "":
                    ret["comment"] = "The following properties were not updated:"
                ret["comment"] = "{} {}".format(ret["comment"], prop)

        ## NOTE: update comment
        if ret["result"] and name in ret["changes"]:
            ret["comment"] = "{} {} was updated".format(dataset_type, name)
        elif ret["result"]:
            ret["comment"] = "{} {} is uptodate".format(dataset_type, name)
        else:
            ret["comment"] = "{} {} failed to be updated".format(dataset_type, name)

    ## NOTE: create or clone the dataset
    elif not exists:
        mod_res_action = "cloned" if cloned_from else "created"
        if __opts__["test"]:
            ## NOTE: pretend to create/clone
            mod_res = OrderedDict([(mod_res_action, True)])
        elif cloned_from:
            ## NOTE: add volsize to properties
            if volume_size:
                properties["volsize"] = volume_size

            ## NOTE: clone the dataset
            mod_res = __salt__["zfs.clone"](
                cloned_from,
                name,
                **{"create_parent": create_parent, "properties": properties}
            )
        else:
            ## NOTE: create the dataset
            mod_res = __salt__["zfs.create"](
                name,
                **{
                    "create_parent": create_parent,
                    "properties": properties,
                    "volume_size": volume_size,
                    "sparse": sparse,
                }
            )

        ret["result"] = mod_res[mod_res_action]
        if ret["result"]:
            ret["changes"][name] = mod_res_action
            if properties:
                ret["changes"][name] = properties
            ret["comment"] = "{} {} was {}".format(
                dataset_type,
                name,
                mod_res_action,
            )
        else:
            ret["comment"] = "failed to {} {} {}".format(
                mod_res_action[:-1],
                dataset_type,
                name,
            )
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]

    return ret


def filesystem_present(name, create_parent=False, properties=None, cloned_from=None):
    """
    ensure filesystem exists and has properties set

    name : string
        name of filesystem
    create_parent : boolean
        creates all the non-existing parent datasets.
        any property specified on the command line using the -o option is ignored.
    cloned_from : string
        name of snapshot to clone
    properties : dict
        additional zfs properties (-o)

    .. note::
        ``cloned_from`` is only use if the filesystem does not exist yet,
        when ``cloned_from`` is set after the filesystem exists it will be ignored.

    .. note::
        Properties do not get cloned, if you specify the properties in the
        state file they will be applied on a subsequent run.

    """
    return _dataset_present(
        "filesystem",
        name,
        properties,
        create_parent=create_parent,
        cloned_from=cloned_from,
    )


def volume_present(
    name,
    volume_size,
    sparse=False,
    create_parent=False,
    properties=None,
    cloned_from=None,
):
    """
    ensure volume exists and has properties set

    name : string
        name of volume
    volume_size : string
        size of volume
    sparse : boolean
        create sparse volume
    create_parent : boolean
        creates all the non-existing parent datasets.
        any property specified on the command line using the -o option is ignored.
    cloned_from : string
        name of snapshot to clone
    properties : dict
        additional zfs properties (-o)

    .. note::
        ``cloned_from`` is only use if the volume does not exist yet,
        when ``cloned_from`` is set after the volume exists it will be ignored.

    .. note::
        Properties do not get cloned, if you specify the properties in the state file
        they will be applied on a subsequent run.

        ``volume_size`` is considered a property, so the volume's size will be
        corrected when the properties get updated if it differs from the
        original volume.

        The sparse parameter is ignored when using ``cloned_from``.

    """
    return _dataset_present(
        "volume",
        name,
        properties,
        volume_size,
        sparse=sparse,
        create_parent=create_parent,
        cloned_from=cloned_from,
    )


def bookmark_present(name, snapshot):
    """
    ensure bookmark exists

    name : string
        name of bookmark
    snapshot : string
        name of snapshot

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## log configuration
    log.debug("zfs.bookmark_present::%s::config::snapshot = %s", name, snapshot)

    ## check we have valid snapshot/bookmark name
    if not __utils__["zfs.is_snapshot"](snapshot):
        ret["result"] = False
        ret["comment"] = "invalid snapshot name: {}".format(name)
        return ret

    if "#" not in name and "/" not in name:
        ## NOTE: simple snapshot name
        #        take the snapshot name and replace the snapshot but with the simple name
        #        e.g. pool/fs@snap + bm --> pool/fs#bm
        name = "{}#{}".format(snapshot[: snapshot.index("@")], name)
        ret["name"] = name

    if not __utils__["zfs.is_bookmark"](name):
        ret["result"] = False
        ret["comment"] = "invalid bookmark name: {}".format(name)
        return ret

    ## ensure bookmark exists
    if not __salt__["zfs.exists"](name, **{"type": "bookmark"}):
        ## NOTE: bookmark the snapshot
        if not __opts__["test"]:
            mod_res = __salt__["zfs.bookmark"](snapshot, name)
        else:
            mod_res = OrderedDict([("bookmarked", True)])

        ret["result"] = mod_res["bookmarked"]
        if ret["result"]:
            ret["changes"][name] = snapshot
            ret["comment"] = "{} bookmarked as {}".format(snapshot, name)
        else:
            ret["comment"] = "failed to bookmark {}".format(snapshot)
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]
    else:
        ## NOTE: bookmark already exists
        ret["comment"] = "bookmark is present"

    return ret


def snapshot_present(name, recursive=False, properties=None):
    """
    ensure snapshot exists and has properties set

    name : string
        name of snapshot
    recursive : boolean
        recursively create snapshots of all descendent datasets
    properties : dict
        additional zfs properties (-o)

    .. note:
        Properties are only set at creation time

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## log configuration
    log.debug("zfs.snapshot_present::%s::config::recursive = %s", name, recursive)
    log.debug("zfs.snapshot_present::%s::config::properties = %s", name, properties)

    ## ensure properties are zfs values
    if properties:
        properties = __utils__["zfs.from_auto_dict"](properties)

    ## check we have valid snapshot name
    if not __utils__["zfs.is_snapshot"](name):
        ret["result"] = False
        ret["comment"] = "invalid snapshot name: {}".format(name)
        return ret

    ## ensure snapshot exits
    if not __salt__["zfs.exists"](name, **{"type": "snapshot"}):
        ## NOTE: create the snapshot
        if not __opts__["test"]:
            mod_res = __salt__["zfs.snapshot"](
                name, **{"recursive": recursive, "properties": properties}
            )
        else:
            mod_res = OrderedDict([("snapshotted", True)])

        ret["result"] = mod_res["snapshotted"]
        if ret["result"]:
            ret["changes"][name] = "snapshotted"
            if properties:
                ret["changes"][name] = properties
            ret["comment"] = "snapshot {} was created".format(name)
        else:
            ret["comment"] = "failed to create snapshot {}".format(name)
            if "error" in mod_res:
                ret["comment"] = mod_res["error"]
    else:
        ## NOTE: snapshot already exists
        ret["comment"] = "snapshot is present"

    return ret


def promoted(name):
    """
    ensure a dataset is not a clone

    name : string
        name of fileset or volume

    .. warning::

        only one dataset can be the origin,
        if you promote a clone the original will now point to the promoted dataset

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## check we if we have a valid dataset name
    if not __utils__["zfs.is_dataset"](name):
        ret["result"] = False
        ret["comment"] = "invalid dataset name: {}".format(name)
        return ret

    ## ensure dataset is the primary instance
    if not __salt__["zfs.exists"](name, **{"type": "filesystem,volume"}):
        ## NOTE: we don't have a dataset
        ret["result"] = False
        ret["comment"] = "dataset {} does not exist".format(name)
    else:
        ## NOTE: check if we have a blank origin (-)
        if (
            __salt__["zfs.get"](
                name, **{"properties": "origin", "fields": "value", "parsable": True}
            )[name]["origin"]["value"]
            == "-"
        ):
            ## NOTE: we're already promoted
            ret["comment"] = "{} already promoted".format(name)
        else:
            ## NOTE: promote dataset
            if not __opts__["test"]:
                mod_res = __salt__["zfs.promote"](name)
            else:
                mod_res = OrderedDict([("promoted", True)])

            ret["result"] = mod_res["promoted"]
            if ret["result"]:
                ret["changes"][name] = "promoted"
                ret["comment"] = "{} promoted".format(name)
            else:
                ret["comment"] = "failed to promote {}".format(name)
                if "error" in mod_res:
                    ret["comment"] = mod_res["error"]

    return ret


def _schedule_snapshot_retrieve(dataset, prefix, snapshots):
    """
    Update snapshots dict with current snapshots

    dataset: string
        name of filesystem or volume
    prefix : string
        prefix for the snapshots
        e.g. 'test' will result in snapshots being named 'test-yyyymmdd_hhmm'
    snapshots : OrderedDict
        preseeded OrderedDict with configuration

    """
    ## NOTE: retrieve all snapshots for the dataset
    for snap in sorted(
        __salt__["zfs.list"](
            dataset, **{"recursive": True, "depth": 1, "type": "snapshot"}
        ).keys()
    ):
        ## NOTE: we only want the actualy name
        ##       myzpool/data@zbck-20171201_000248 -> zbck-20171201_000248
        snap_name = snap[snap.index("@") + 1 :]

        ## NOTE: we only want snapshots matching our prefix
        if not snap_name.startswith("{}-".format(prefix)):
            continue

        ## NOTE: retrieve the holds for this snapshot
        snap_holds = __salt__["zfs.holds"](snap)

        ## NOTE: this snapshot has no holds, eligable for pruning
        if not snap_holds:
            snapshots["_prunable"].append(snap)

        ## NOTE: update snapshots based on holds (if any)
        ##       we are only interested in the ones from our schedule
        ##       if we find any others we skip them
        for hold in snap_holds:
            if hold in snapshots["_schedule"].keys():
                snapshots[hold].append(snap)

    return snapshots


def _schedule_snapshot_prepare(dataset, prefix, snapshots):
    """
    Update snapshots dict with info for a new snapshot

    dataset: string
        name of filesystem or volume
    prefix : string
        prefix for the snapshots
        e.g. 'test' will result in snapshots being named 'test-yyyymmdd_hhmm'
    snapshots : OrderedDict
        preseeded OrderedDict with configuration

    """
    ## NOTE: generate new snapshot name
    snapshot_create_name = "{dataset}@{prefix}-{timestamp}".format(
        dataset=dataset,
        prefix=prefix,
        timestamp=datetime.now().strftime("%Y%m%d_%H%M%S"),
    )

    ## NOTE: figure out if we need to create the snapshot
    timestamp_now = datetime.now().replace(second=0, microsecond=0)
    snapshots["_create"][snapshot_create_name] = []
    for hold, hold_count in snapshots["_schedule"].items():
        ## NOTE: skip hold if we don't keep snapshots for it
        if hold_count == 0:
            continue

        ## NOTE: figure out if we need the current hold on the new snapshot
        if snapshots[hold]:
            ## NOTE: extract datetime from snapshot name
            timestamp = datetime.strptime(
                snapshots[hold][-1],
                "{}@{}-%Y%m%d_%H%M%S".format(dataset, prefix),
            ).replace(second=0, microsecond=0)

            ## NOTE: compare current timestamp to timestamp from snapshot
            if hold == "minute" and timestamp_now <= timestamp:
                continue
            elif hold == "hour" and timestamp_now.replace(
                **comp_hour
            ) <= timestamp.replace(**comp_hour):
                continue
            elif hold == "day" and timestamp_now.replace(
                **comp_day
            ) <= timestamp.replace(**comp_day):
                continue
            elif hold == "month" and timestamp_now.replace(
                **comp_month
            ) <= timestamp.replace(**comp_month):
                continue
            elif hold == "year" and timestamp_now.replace(
                **comp_year
            ) <= timestamp.replace(**comp_year):
                continue

        ## NOTE: add hold entry for snapshot
        snapshots["_create"][snapshot_create_name].append(hold)

    return snapshots


def scheduled_snapshot(name, prefix, recursive=True, schedule=None):
    """
    maintain a set of snapshots based on a schedule

    name : string
        name of filesystem or volume
    prefix : string
        prefix for the snapshots
        e.g. 'test' will result in snapshots being named 'test-yyyymmdd_hhmm'
    recursive : boolean
        create snapshots for all children also
    schedule : dict
        dict holding the schedule, the following keys are available (minute, hour,
        day, month, and year) by default all are set to 0 the value indicated the
        number of snapshots of that type to keep around.

    .. warning::

        snapshots will only be created and pruned every time the state runs.
        a schedule must be setup to automatically run the state. this means that if
        you run the state daily the hourly snapshot will only be made once per day!

    .. versionchanged:: 2018.3.0

        switched to localtime from gmtime so times now take into account timezones.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    ## initialize defaults
    schedule_holds = ["minute", "hour", "day", "month", "year"]
    snapshots = OrderedDict(
        [("_create", OrderedDict()), ("_prunable", []), ("_schedule", OrderedDict())]
    )

    ## strict configuration validation
    ## NOTE: we need a valid dataset
    if not __utils__["zfs.is_dataset"](name):
        ret["result"] = False
        ret["comment"] = "invalid dataset name: {}".format(name)

    if not __salt__["zfs.exists"](name, **{"type": "filesystem,volume"}):
        ret["comment"] = "dataset {} does not exist".format(name)
        ret["result"] = False

    ## NOTE: prefix must be 4 or longer
    if not prefix or len(prefix) < 4:
        ret["comment"] = "prefix ({}) must be at least 4 long".format(prefix)
        ret["result"] = False

    ## NOTE: validate schedule
    total_count = 0
    for hold in schedule_holds:
        snapshots[hold] = []
        if hold not in schedule:
            snapshots["_schedule"][hold] = 0
        elif isinstance(schedule[hold], int):
            snapshots["_schedule"][hold] = schedule[hold]
        else:
            ret["result"] = False
            ret["comment"] = "schedule value for {} is not an integer".format(
                hold,
            )
            break
        total_count += snapshots["_schedule"][hold]
    if ret["result"] and total_count == 0:
        ret["result"] = False
        ret["comment"] = "schedule is not valid, you need to keep atleast 1 snapshot"

    ## NOTE: return if configuration is not valid
    if not ret["result"]:
        return ret

    ## retrieve existing snapshots
    snapshots = _schedule_snapshot_retrieve(name, prefix, snapshots)

    ## prepare snapshot
    snapshots = _schedule_snapshot_prepare(name, prefix, snapshots)

    ## log configuration
    log.debug("zfs.scheduled_snapshot::%s::config::recursive = %s", name, recursive)
    log.debug("zfs.scheduled_snapshot::%s::config::prefix = %s", name, prefix)
    log.debug("zfs.scheduled_snapshot::%s::snapshots = %s", name, snapshots)

    ## create snapshot(s)
    for snapshot_name, snapshot_holds in snapshots["_create"].items():
        ## NOTE: skip if new snapshot has no holds
        if not snapshot_holds:
            continue

        ## NOTE: create snapshot
        if not __opts__["test"]:
            mod_res = __salt__["zfs.snapshot"](
                snapshot_name, **{"recursive": recursive}
            )
        else:
            mod_res = OrderedDict([("snapshotted", True)])

        if not mod_res["snapshotted"]:
            ret["result"] = False
            ret["comment"] = "error creating snapshot ({})".format(snapshot_name)
        else:
            ## NOTE: create holds (if we have a snapshot)
            for hold in snapshot_holds:
                if not __opts__["test"]:
                    mod_res = __salt__["zfs.hold"](
                        hold, snapshot_name, **{"recursive": recursive}
                    )
                else:
                    mod_res = OrderedDict([("held", True)])

                if not mod_res["held"]:
                    ret["result"] = False
                    ret["comment"] = "error adding hold ({}) to snapshot ({})".format(
                        hold,
                        snapshot_name,
                    )
                    break

                snapshots[hold].append(snapshot_name)

        if ret["result"]:
            ret["comment"] = "scheduled snapshots updated"
            if "created" not in ret["changes"]:
                ret["changes"]["created"] = []
            ret["changes"]["created"].append(snapshot_name)

    ## prune hold(s)
    for hold, hold_count in snapshots["_schedule"].items():
        while ret["result"] and len(snapshots[hold]) > hold_count:
            ## NOTE: pop oldest snapshot
            snapshot_name = snapshots[hold].pop(0)

            ## NOTE: release hold for snapshot
            if not __opts__["test"]:
                mod_res = __salt__["zfs.release"](
                    hold, snapshot_name, **{"recursive": recursive}
                )
            else:
                mod_res = OrderedDict([("released", True)])

            if not mod_res["released"]:
                ret["result"] = False
                ret["comment"] = "error adding hold ({}) to snapshot ({})".format(
                    hold,
                    snapshot_name,
                )

            ## NOTE: mark as prunable
            if not __salt__["zfs.holds"](snapshot_name):
                snapshots["_prunable"].append(snapshot_name)

    ## prune snapshot(s)
    for snapshot_name in snapshots["_prunable"]:
        ## NOTE: destroy snapshot
        if not __opts__["test"]:
            mod_res = __salt__["zfs.destroy"](snapshot_name, **{"recursive": recursive})
        else:
            mod_res = OrderedDict([("destroyed", True)])

        if not mod_res["destroyed"]:
            ret["result"] = False
            ret["comment"] = "error prunding snapshot ({1})".format(
                snapshot_name,
            )
            break

    if ret["result"] and snapshots["_prunable"]:
        ret["comment"] = "scheduled snapshots updated"
        ret["changes"]["pruned"] = snapshots["_prunable"]

    if ret["result"] and not ret["changes"]:
        ret["comment"] = "scheduled snapshots are up to date"

    return ret



import salt.modules.linux_acl as linux_acl
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LinuxAclTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        self.cmdrun = MagicMock()
        self.addCleanup(delattr, self, "cmdrun")
        return {linux_acl: {"__salt__": {"cmd.run": self.cmdrun}}}

    def setUp(self):
        self.file = "/tmp/file"
        self.quoted_file = '"/tmp/file"'
        self.files = ["/tmp/file1", "/tmp/file2", "/tmp/file3 with whitespaces"]
        self.quoted_files = ['"{}"'.format(f) for f in self.files]
        self.u_acl = ["u", "myuser", "rwx"]
        self.user_acl = ["user", "myuser", "rwx"]
        self.user_acl_cmd = "u:myuser:rwx"
        self.g_acl = ["g", "mygroup", "rwx"]
        self.group_acl = ["group", "mygroup", "rwx"]
        self.group_acl_cmd = "g:mygroup:rwx"
        self.d_u_acl = ["d:u", "myuser", "rwx"]
        self.d_user_acl = ["d:user", "myuser", "rwx"]
        self.default_user_acl = ["d:user", "myuser", "rwx"]
        self.default_user_acl_cmd = "d:u:myuser:rwx"

    def tearDown(self):
        for attrname in (
            "file",
            "quoted_file",
            "files",
            "quoted_files",
            "u_acl",
            "user_acl",
            "user_acl_cmd",
            "g_acl",
            "group_acl",
            "group_acl_cmd",
            "d_u_acl",
            "d_user_acl",
            "default_user_acl",
            "default_user_acl_cmd",
        ):
            delattr(self, attrname)

    # too easy to test (DRY)
    def test_version(self):
        pass

    def test_getfacl_wo_args(self):
        self.assertRaises(CommandExecutionError, linux_acl.getfacl)

    def test_getfacl_w_single_arg(self):
        linux_acl.getfacl(self.file)
        self.cmdrun.assert_called_once_with(
            "getfacl --absolute-names " + self.quoted_file, python_shell=False
        )

    def test_getfacl_w_multiple_args(self):
        linux_acl.getfacl(*self.files)
        self.cmdrun.assert_called_once_with(
            "getfacl --absolute-names " + " ".join(self.quoted_files),
            python_shell=False,
        )

    def test_getfacl__recursive_w_multiple_args(self):
        linux_acl.getfacl(*self.files, recursive=True)
        self.cmdrun.assert_called_once_with(
            "getfacl --absolute-names -R " + " ".join(self.quoted_files),
            python_shell=False,
        )

    def test_getfacl__effective_acls(self):
        line = "group:webmaster:r-x        #effective:---"
        user = "root"
        group = "root"
        expected = {
            "type": "acl",
            "group": "webmaster",
            "permissions": {"read": False, "write": False, "execute": False},
            "octal": 0,
        }
        self.assertEqual(linux_acl._parse_acl(line, user, group), expected)

    def test_wipefacls_wo_args(self):
        self.assertRaises(CommandExecutionError, linux_acl.wipefacls)

    def test_wipefacls_w_single_arg(self):
        linux_acl.wipefacls(self.file)
        self.cmdrun.assert_called_once_with(
            "setfacl -b " + self.quoted_file, python_shell=False
        )

    def test_wipefacls_w_multiple_args(self):
        linux_acl.wipefacls(*self.files)
        self.cmdrun.assert_called_once_with(
            "setfacl -b " + " ".join(self.quoted_files), python_shell=False
        )

    def test_wipefacls__recursive_w_multiple_args(self):
        linux_acl.wipefacls(*self.files, recursive=True)
        self.cmdrun.assert_called_once_with(
            "setfacl -b -R " + " ".join(self.quoted_files), python_shell=False
        )

    def test_modfacl_wo_args(self):
        for acl in [self.u_acl, self.user_acl, self.g_acl, self.group_acl]:
            self.assertRaises(CommandExecutionError, linux_acl.modfacl, *acl)

    def test_modfacl__u_w_single_arg(self):
        linux_acl.modfacl(*(self.u_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.user_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__u_w_multiple_args(self):
        linux_acl.modfacl(*(self.u_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__user_w_single_arg(self):
        linux_acl.modfacl(*(self.user_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.user_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__user_w_multiple_args(self):
        linux_acl.modfacl(*(self.user_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__g_w_single_arg(self):
        linux_acl.modfacl(*(self.g_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.group_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__g_w_multiple_args(self):
        linux_acl.modfacl(*(self.g_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.group_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__group_w_single_arg(self):
        linux_acl.modfacl(*(self.group_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.group_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__group_w_multiple_args(self):
        linux_acl.modfacl(*(self.group_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.group_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__d_u_w_single_arg(self):
        linux_acl.modfacl(*(self.d_u_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__d_u_w_multiple_args(self):
        linux_acl.modfacl(*(self.d_u_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__d_user_w_single_arg(self):
        linux_acl.modfacl(*(self.d_user_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__d_user_w_multiple_args(self):
        linux_acl.modfacl(*(self.d_user_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__default_user_w_single_arg(self):
        linux_acl.modfacl(*(self.default_user_acl + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd, self.quoted_file]),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__default_user_w_multiple_args(self):
        linux_acl.modfacl(*(self.default_user_acl + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -m " + " ".join([self.default_user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl__recursive_w_multiple_args(self):
        linux_acl.modfacl(*(self.user_acl + self.files), recursive=True)
        self.cmdrun.assert_called_once_with(
            "setfacl -R -m " + " ".join([self.user_acl_cmd] + self.quoted_files),
            python_shell=False,
            raise_err=False,
        )

    def test_modfacl_raise_err(self):
        mock = MagicMock(side_effect=CommandExecutionError("Custom err"))
        with patch.dict(linux_acl.__salt__, {"cmd.run": mock}):
            with self.assertRaises(CommandExecutionError) as excinfo:
                linux_acl.modfacl(*(self.user_acl + self.files), raise_err=True)
            self.assertEqual(excinfo.exception.strerror, "Custom err")

    def test_delfacl_wo_args(self):
        for acl in [self.u_acl, self.user_acl, self.g_acl, self.group_acl]:
            self.assertRaises(CommandExecutionError, linux_acl.delfacl, *acl[:-1])

    def test_delfacl__u_w_single_arg(self):
        linux_acl.delfacl(*(self.u_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.user_acl_cmd.rpartition(":")[0], self.quoted_file]),
            python_shell=False,
        )

    def test_delfacl__u_w_multiple_args(self):
        linux_acl.delfacl(*(self.u_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.user_acl_cmd.rpartition(":")[0]] + self.quoted_files),
            python_shell=False,
        )

    def test_delfacl__user_w_single_arg(self):
        linux_acl.delfacl(*(self.user_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.user_acl_cmd.rpartition(":")[0], self.quoted_file]),
            python_shell=False,
        )

    def test_delfacl__user_w_multiple_args(self):
        linux_acl.delfacl(*(self.user_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.user_acl_cmd.rpartition(":")[0]] + self.quoted_files),
            python_shell=False,
        )

    def test_delfacl__g_w_single_arg(self):
        linux_acl.delfacl(*(self.g_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.group_acl_cmd.rpartition(":")[0], self.quoted_file]),
            python_shell=False,
        )

    def test_delfacl__g_w_multiple_args(self):
        linux_acl.delfacl(*(self.g_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.group_acl_cmd.rpartition(":")[0]] + self.quoted_files),
            python_shell=False,
        )

    def test_delfacl__group_w_single_arg(self):
        linux_acl.delfacl(*(self.group_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.group_acl_cmd.rpartition(":")[0], self.quoted_file]),
            python_shell=False,
        )

    def test_delfacl__group_w_multiple_args(self):
        linux_acl.delfacl(*(self.group_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join([self.group_acl_cmd.rpartition(":")[0]] + self.quoted_files),
            python_shell=False,
        )

    def test_delfacl__d_u_w_single_arg(self):
        linux_acl.delfacl(*(self.d_u_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0], self.quoted_file]
            ),
            python_shell=False,
        )

    def test_delfacl__d_u_w_multiple_args(self):
        linux_acl.delfacl(*(self.d_u_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0]] + self.quoted_files
            ),
            python_shell=False,
        )

    def test_delfacl__d_user_w_single_arg(self):
        linux_acl.delfacl(*(self.d_user_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0], self.quoted_file]
            ),
            python_shell=False,
        )

    def test_delfacl__d_user_w_multiple_args(self):
        linux_acl.delfacl(*(self.d_user_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0]] + self.quoted_files
            ),
            python_shell=False,
        )

    def test_delfacl__default_user_w_single_arg(self):
        linux_acl.delfacl(*(self.default_user_acl[:-1] + [self.file]))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0], self.quoted_file]
            ),
            python_shell=False,
        )

    def test_delfacl__default_user_w_multiple_args(self):
        linux_acl.delfacl(*(self.default_user_acl[:-1] + self.files))
        self.cmdrun.assert_called_once_with(
            "setfacl -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0]] + self.quoted_files
            ),
            python_shell=False,
        )

    def test_delfacl__recursive_w_multiple_args(self):
        linux_acl.delfacl(*(self.default_user_acl[:-1] + self.files), recursive=True)
        self.cmdrun.assert_called_once_with(
            "setfacl -R -x "
            + " ".join(
                [self.default_user_acl_cmd.rpartition(":")[0]] + self.quoted_files
            ),
            python_shell=False,
        )
